import os
import sys
import tempfile
from weakref import WeakValueDictionary
import dbfile
import logger


class FilePopenBase(object):
    '''Base class for subprocess.Popen-like class interface that
can be supported on Python 2.3 (without subprocess).  The main goal
is to avoid the pitfalls of Popen.communicate(), which cannot handle
more than a small amount of data, and to avoid the possibility of deadlocks
and the issue of threading, by using temporary files'''

    def __init__(self, args, bufsize=0, executable=None,
                 stdin=None, stdout=None, stderr=None, *largs, **kwargs):
        '''Mimics the interface of subprocess.Popen() with two additions:
- stdinFlag, if passed, gives a flag to add the stdin filename directly
to the command line (rather than passing it by redirecting stdin).
example: stdinFlag="-i" will add the following to the commandline:
-i /path/to/the/file
If set to None, the stdin filename is still appended to the commandline,
but without a preceding flag argument.

-stdoutFlag: exactly the same thing, except for the stdout filename.
'''
        self.stdin, self._close_stdin = self._get_pipe_file(stdin, 'stdin')
        self.stdout, self._close_stdout = self._get_pipe_file(stdout, 'stdout')
        self.stderr, self._close_stderr = self._get_pipe_file(stderr, 'stderr')
        kwargs = kwargs.copy() # get a copy we can change
        try: # add as a file argument
            stdinFlag = kwargs['stdinFlag']
            if stdinFlag:
                args.append(stdinFlag)
            args.append(self._stdin_path)
            del kwargs['stdinFlag']
            stdin = None
        except KeyError: # just make it read this stream
            stdin = self.stdin
        try: # add as a file argument
            stdoutFlag = kwargs['stdoutFlag']
            if stdoutFlag:
                args.append(stdoutFlag)
            args.append(self._stdout_path)
            del kwargs['stdoutFlag']
            stdout = None
        except KeyError: # make it write to this stream
            stdout = self.stdout
        self.args = (args, bufsize, executable, stdin, stdout,
                     self.stderr) + largs
        self.kwargs = kwargs

    def _get_pipe_file(self, ifile, attr):
        'create a temp filename instead of a PIPE; save the filename'
        if ifile == PIPE: # use temp file instead!
            fd, path = tempfile.mkstemp()
            setattr(self, '_' + attr + '_path', path)
            return os.fdopen(fd, 'w+b'), True
        elif ifile is not None:
            setattr(self, '_' + attr + '_path', ifile.name)
        return ifile, False

    def _close_file(self, attr):
        'close and delete this temp file if still open'
        if getattr(self, '_close_' + attr):
            getattr(self, attr).close()
            setattr(self, '_close_' + attr, False)
            os.remove(getattr(self, '_' + attr + '_path'))

    def _rewind_for_reading(self, ifile):
        if ifile is not None:
            ifile.flush()
            ifile.seek(0)

    def close(self):
        """Close any open temp (PIPE) files. """
        self._close_file('stdin')
        self._close_file('stdout')
        self._close_file('stderr')

    def __del__(self):
        self.close()


def call_subprocess(*popenargs, **kwargs):
    'portable interface to subprocess.call(), even if subprocess not available'
    p = FilePopen(*popenargs, **kwargs)
    return p.wait()

try:
    import subprocess
    PIPE = subprocess.PIPE

    class FilePopen(FilePopenBase):
        'this subclass uses the subprocess module Popen() functionality'

        def wait(self):
            self._rewind_for_reading(self.stdin)
            p = subprocess.Popen(*self.args, **self.kwargs)
            p.wait()
            self._close_file('stdin')
            self._rewind_for_reading(self.stdout)
            self._rewind_for_reading(self.stderr)
            return p.returncode

except ImportError:
    CSH_REDIRECT = False # SH style redirection is default
    import platform
    if platform.system() == 'Windows':

        def mkarg(arg):
            """Very basic quoting of arguments for Windows """
            return '"' + arg + '"'
    else: # UNIX
        from commands import mkarg
        try:
            if os.environ['SHELL'].endswith('csh'):
                CSH_REDIRECT = True
        except KeyError:
            pass

    badExecutableCode = None

    class FilePopen(FilePopenBase):
        'this subclass fakes subprocess.Popen.wait() using os.system()'

        def wait(self):
            self._rewind_for_reading(self.stdin)
            args = map(mkarg, self.args[0])
            if self.args[3]: # redirect stdin
                args += ['<', mkarg(self._stdin_path)]
            if self.args[4]: # redirect stdout
                args += ['>', mkarg(self._stdout_path)]
            cmd = ' '.join(args)
            if self.args[5]: # redirect stderr
                if CSH_REDIRECT:
                    cmd = '(%s) >& %s' % (cmd, mkarg(self._stderr_path))
                else:
                    cmd = cmd + ' 2> ' + mkarg(self._stderr_path)
            returncode = os.system(cmd)
            self._close_file('stdin')
            self._rewind_for_reading(self.stdout)
            self._rewind_for_reading(self.stderr)
            if badExecutableCode is not None and \
               badExecutableCode == returncode:
                raise OSError('no such command: %s' % str(self.args[0]))
            return returncode
    PIPE = id(FilePopen) # an arbitrary code for identifying this code

    # find out exit code for a bad executable name, silently
    badExecutableCode = call_subprocess(('aZfDqW9', ),
                                        stdout=PIPE, stderr=PIPE)


def ClassicUnpickler(cls, state):
    'standard python unpickling behavior'
    self = cls.__new__(cls)
    try:
        setstate = self.__setstate__
    except AttributeError: # JUST SAVE TO __dict__ AS USUAL
        self.__dict__.update(state)
    else:
        setstate(state)
    return self

ClassicUnpickler.__safe_for_unpickling__ = 1


def filename_unpickler(cls, path, kwargs):
    'raise IOError if path not readable'
    if not os.path.exists(path):
        try:# CONVERT TO ABSOLUTE PATH BASED ON SAVED DIRECTORY PATH
            path = os.path.normpath(os.path.join(kwargs['curdir'], path))
            if not os.path.exists(path):
                raise IOError('unable to open file %s' % path)
        except KeyError:
            raise IOError('unable to open file %s' % path)
    if cls is SourceFileName:
        return SourceFileName(path)
    raise ValueError('Attempt to unpickle untrusted class ' + cls.__name__)

filename_unpickler.__safe_for_unpickling__ = 1


class SourceFileName(str):
    '''store a filepath string, raise IOError on unpickling
if filepath not readable, and complain if the user tries
to pickle a relative path'''

    def __reduce__(self):
        if not os.path.isabs(str(self)):
            print >>sys.stderr, '''
WARNING: You are trying to pickle an object that has a local
file dependency stored only as a relative file path:
%s
This is not a good idea, because anyone working in
a different directory would be unable to unpickle this object,
since it would be unable to find the file using the relative path.
To avoid this problem, SourceFileName is saving the current
working directory path so that it can translate the relative
path to an absolute path.  In the future, please use absolute
paths when constructing objects that you intend to save to worldbase
or pickle!''' % str(self)
        return (filename_unpickler, (self.__class__, str(self),
                                    dict(curdir=os.getcwd())))


def file_dirpath(filename):
    'return path to directory containing filename'
    dirname = os.path.dirname(filename)
    if dirname == '':
        return os.curdir
    else:
        return dirname


def get_valid_path(*pathTuples):
    '''for each tuple in args, build path using os.path.join(),
    and return the first path that actually exists, or else None.'''
    for t in pathTuples:
        mypath = os.path.join(*t)
        if os.path.exists(mypath):
            return mypath


def search_dirs_for_file(filepath, pathlist=()):
    'return successful path based on trying pathlist locations'
    if os.path.exists(filepath):
        return filepath
    b = os.path.basename(filepath)
    for s in pathlist: # NOW TRY EACH DIRECTORY IN pathlist
        mypath = os.path.join(s, b)
        if os.path.exists(mypath):
            return mypath
    raise IOError('unable to open %s from any location in %s'
                  % (filepath, pathlist))


def report_exception():
    'print string message from exception to stderr'
    import traceback
    info = sys.exc_info()[:2]
    l = traceback.format_exception_only(info[0], info[1])
    print >>sys.stderr, 'Warning: caught %s\nContinuing...' % l[0]


def standard_invert(self):
    'keep a reference to an inverse mapping, using self._inverseClass'
    try:
        return self._inverse
    except AttributeError:
        self._inverse = self._inverseClass(self)
        return self._inverse


def lazy_create_invert(klass):
    """Create a function to replace __invert__ with a call to a cached object.

    lazy_create_invert defines a method that looks up self._inverseObj
    and, it it doesn't exist, creates it from 'klass' and then saves it.
    The resulting object is then returned as the inverse.  This allows
    for one-time lazy creation of a single object per parent class.
    """

    def invert_fn(self, klass=klass):
        try:
            return self._inverse
        except AttributeError:
            # does not exist yet; create & store.
            inverseObj = klass(self)
            self._inverse = inverseObj
            return inverseObj

    return invert_fn


def standard_getstate(self):
    'get dict of attributes to save, using self._pickleAttrs dictionary'
    d = {}
    for attr, arg in self._pickleAttrs.items():
        try:
            if isinstance(arg, str):
                d[arg] = getattr(self, attr)
            else:
                d[attr] = getattr(self, attr)
        except AttributeError:
            pass
    try:
        # DON'T SAVE itemClass IF SIMPLY A SHADOW of default itemClass
        # from __class__
        if not hasattr(self.__class__, 'itemClass') or \
           (self.itemClass is not self.__class__.itemClass and
            (not hasattr(self.itemClass, '_shadowParent') or
             self.itemClass._shadowParent is not self.__class__.itemClass)):
            try:
                d['itemClass'] = self.itemClass._shadowParent
            except AttributeError:
                d['itemClass'] = self.itemClass
        if not hasattr(self.__class__, 'itemSliceClass') or \
           (self.itemSliceClass is not self.__class__.itemSliceClass and
            (not hasattr(self.itemSliceClass, '_shadowParent') or
             self.itemSliceClass._shadowParent is not
             self.__class__.itemSliceClass)):
            try:
                d['itemSliceClass'] = self.itemSliceClass._shadowParent
            except AttributeError:
                d['itemSliceClass'] = self.itemSliceClass
    except AttributeError:
        pass
    try: # SAVE CUSTOM UNPACKING METHOD
        d['unpack_edge'] = self.__dict__['unpack_edge']
    except KeyError:
        pass
    return d


def standard_setstate(self, state):
    'apply dict of saved state by passing as kwargs to constructor'
    if isinstance(state, list):  # GET RID OF THIS BACKWARDS-COMPATIBILITY CODE
        self.__init__(*state)
        print >>sys.stderr, 'WARNING: obsolete list pickle %s. Update \
                by resaving!' % repr(self)
    else:
        state['unpicklingMode'] = True # SIGNAL THAT WE ARE UNPICKLING
        self.__init__(**state)


def apply_itemclass(self, state):
    try:
        self.itemClass = state['itemClass']
        self.itemSliceClass = state['itemSliceClass']
    except KeyError:
        pass


def generate_items(items):
    'generate o.id,o for o in items'
    for o in items:
        yield o.id, o


def item_unpickler(db, *args):
    'get an item or subslice of a database'
    obj = db
    for arg in args:
        obj = obj[arg]
    return obj
item_unpickler.__safe_for_unpickling__ = 1


def item_reducer(self): ############################# SUPPORT FOR PICKLING
    'pickle an item of a database just as a reference'
    return (item_unpickler, (self.db, self.id))


def shadow_reducer(self):
    'pickle shadow class instance using its true class'
    shadowClass = self.__class__
    trueClass = shadowClass._shadowParent # super() TOTALLY FAILED ME HERE!
    self.__class__ = trueClass # FORCE IT TO PICKLE USING trueClass
    keepDB = False
    if hasattr(shadowClass, 'db') and not hasattr(self, 'db'):
        keepDB = True
        self.__dict__['db'] = shadowClass.db # retain this attribute!!
    if hasattr(trueClass, '__reduce__'): # USE trueClass.__reduce__
        result = trueClass.__reduce__(self)
    elif hasattr(trueClass, '__getstate__'): # USE trueClass.__getstate__
        result = (ClassicUnpickler, (trueClass, self.__getstate__()))
    else: # PICKLE __dict__ AS USUAL PYTHON PRACTICE
        result = (ClassicUnpickler, (trueClass, self.__dict__))
    self.__class__ = shadowClass # RESTORE SHADOW CLASS
    if keepDB: # now we can drop the temporary db attribute we added
        del self.__dict__['db']
    return result


def get_bound_subclass(obj, classattr='__class__', subname=None, factories=(),
                       attrDict=None, subclassArgs=None):
    'create a subclass specifically for obj to bind its shadow attributes'
    targetClass = getattr(obj, classattr)
    try:
        if targetClass._shadowOwner is obj:
            return targetClass # already shadowed, so nothing to do
    except AttributeError: # not a shadow class, so just shadow it
        pass
    else: # someone else's shadow class, so shadow its parent
        targetClass = targetClass._shadowParent
    if subname is None: # get a name from worldbase ID
        try:
            subname = obj._persistent_id.split('.')[-1]
        except AttributeError:
            subname = '__generic__'

    class shadowClass(targetClass):
        __reduce__ = shadow_reducer
        _shadowParent = targetClass # NEED TO KNOW ORIGINAL CLASS
        _shadowOwner = obj # need to know who owns it
        if attrDict is not None: # add attributes to the class dictionary
            locals().update(attrDict)
        for f in factories:
            f(locals())

    if classattr == 'itemClass' or classattr == 'itemSliceClass':
        shadowClass.db = obj # the class needs to know its db object
    try: # run subclass initializer if present
        subclass_init = shadowClass._init_subclass
    except AttributeError: # no subclass initializer, so nothing to do
        pass
    else: # run the subclass initializer
        if subclassArgs is None:
            subclassArgs = {}
        subclass_init(**subclassArgs)
    shadowClass.__name__ = targetClass.__name__ + '_' + subname
    setattr(obj, classattr, shadowClass) # SHADOW CLASS REPLACES ORIGINAL
    return shadowClass


def method_not_implemented(*args, **kwargs):
    raise NotImplementedError


def read_only_error(*args, **kwargs):
    raise NotImplementedError("read only dict")


def methodFactory(methodList, methodStr, localDict):
    'save a method or exec expression for each name in methodList'
    for methodName in methodList:
        if callable(methodStr):
            localDict[methodName] = methodStr
        else:
            localDict[methodName] = eval(methodStr % methodName)


def open_shelve(filename, mode=None, writeback=False, allowReadOnly=False,
                useHash=False, verbose=True):
    '''Alternative to shelve.open with several benefits:
- uses bsddb btree by default instead of bsddb hash, which is very slow
  for large databases.  Will automatically fall back to using bsddb hash
  for existing hash-based shelve files.  Set useHash=True to force it to use
  bsddb hash.

- allowReadOnly=True will automatically suppress permissions errors so
  user can at least get read-only access to the desired shelve, if no write
  permission.

- mode=None first attempts to open file in read-only mode, but if the file
  does not exist, opens it in create mode.

- raises standard exceptions defined in dbfile: WrongFormatError,
  PermissionsError, ReadOnlyError, NoSuchFileError

- avoids generating bogus __del__ warnings as Python shelve.open() does.
  '''
    if mode=='r': # READ-ONLY MODE, RAISE EXCEPTION IF NOT FOUND
        return dbfile.shelve_open(filename, flag=mode, useHash=useHash)
    elif mode is None:
        try: # 1ST TRY READ-ONLY, BUT IF NOT FOUND CREATE AUTOMATICALLY
            return dbfile.shelve_open(filename, 'r', useHash=useHash)
        except dbfile.NoSuchFileError:
            mode = 'c' # CREATE NEW SHELVE FOR THE USER
    try: # CREATION / WRITING: FORCE IT TO WRITEBACK AT close() IF REQUESTED
        return dbfile.shelve_open(filename, flag=mode, writeback=writeback,
                                  useHash=useHash)
    except dbfile.ReadOnlyError:
        if allowReadOnly:
            d = dbfile.shelve_open(filename, flag='r', useHash=useHash)
            if verbose:
                logger.warn('''
Opening shelve file %s in read-only mode because you lack
write permissions to this file.  You will NOT be able to modify the contents
of this shelve dictionary.  To avoid seeing this warning message,
use verbose=False argument for the classutil.open_shelve() function.
''' % filename)
            return d
        else:
            raise


def get_shelve_or_dict(filename=None, dictClass=None, **kwargs):
    if filename is not None:
        if dictClass is not None:
            return dictClass(filename, **kwargs)
        else:
            from mapping import IntShelve
            return IntShelve(filename, **kwargs)
    return {}


class PathSaver(object):

    def __init__(self, origPath):
        self.origPath = origPath
        self.origDir = os.getcwd()

    def __str__(self):
        if os.access(self.origPath, os.R_OK):
            return self.origPath
        trypath = os.path.join(self.origDir, self.origPath)
        if os.access(trypath, os.R_OK):
            return trypath


def override_rich_cmp(localDict):
    'create rich comparison methods that just use __cmp__'
    mycmp = localDict['__cmp__']
    localDict['__lt__'] = lambda self, other: mycmp(self, other) < 0
    localDict['__le__'] = lambda self, other: mycmp(self, other) <= 0
    localDict['__eq__'] = lambda self, other: mycmp(self, other) == 0
    localDict['__ne__'] = lambda self, other: mycmp(self, other) != 0
    localDict['__gt__'] = lambda self, other: mycmp(self, other) > 0
    localDict['__ge__'] = lambda self, other: mycmp(self, other) >= 0


class DBAttributeDescr(object):
    'obtain an attribute from associated db object'

    def __init__(self, attr):
        self.attr = attr

    def __get__(self, obj, objtype):
        return getattr(obj.db, self.attr)


def get_env_or_cwd(envname):
    'get the specified environment value or path to current directory'
    try:
        return os.environ[envname] # USER-SPECIFIED DIRECTORY
    except KeyError:
        return os.getcwd() # DEFAULT: SAVE IN CURRENT DIRECTORY


class RecentValueDictionary(WeakValueDictionary):
    '''keep the most recent n references in a WeakValueDictionary.
    This combines the elegant cache behavior of a WeakValueDictionary
    (only keep an item in cache if the user is still using it),
    with the most common efficiency pattern: locality, i.e.
    references to a given thing tend to cluster in time.  Note that
    this works *even* if the user is not holding a reference to
    the item between requests for it.  Our Most Recent queue will
    hold a reference to it, keeping it in the WeakValueDictionary,
    until it is bumped by more recent requests.

    n: the maximum number of objects to keep in the Most Recent queue,
       default value 50.'''

    def __init__(self, n=None):
        WeakValueDictionary.__init__(self)
        if n<1: # user doesn't want any Most Recent value queue
            self.__class__ = WeakValueDictionary # revert to regular WVD
            return
        if isinstance(n, int):
            self.n = n # size limit
        else:
            self.n = 50
        self._head = self._tail = None
        self._keepDict = {} # most recent queue

    def __getitem__(self, k):
        v = WeakValueDictionary.__getitem__(self, k) # KeyError if not found
        self.keep_this(v)
        return v

    def _splice(self, previous, after):
        'link previous <--> after in queue, setting head & tail if needed'
        if previous is not None:
            self._keepDict[previous][1] = after
        if after is not None:
            self._keepDict[after][0] = previous
        elif previous is not None: # previous is end of queue!
            self._tail = previous
        if after is self._head:
            self._head = previous
            
    def keep_this(self, v):
        """add v as our most recent ref; drop oldest ref if over size limit.
        """
        if v is self._head:
            return # already at head of queue, so nothing to do
        try: # check if already in _keepDict
            previous, after = self._keepDict[v]
        except KeyError:
            self._keepDict[v] = [None, None]
        else: # remove from current position
            self._splice(previous, after)
            self._keepDict[v][0] = None
        self._splice(v, self._head) # place at head of queue
        if len(self._keepDict) > self.n: # delete oldest entry
            vdel = self._tail # get current tail
            self._splice(self._keepDict[vdel][0], None) # set new tail
            del self._keepDict[vdel]

    def __setitem__(self, k, v):
        WeakValueDictionary.__setitem__(self, k, v)
        self.keep_this(v)

    def clear(self):
        self._head = self._tail = None
        self._keepDict.clear()
        WeakValueDictionary.clear(self)

    def __repr__(self):
        return '<RecentValueDictionary object: %d members, cachesize %d>' %\
               (len(self._keepDict), self.n)


def make_attribute_interface(d):
    """
    If 'd' contains int values, use them to index tuples.

    If 'd' contains str values, use them to retrieve attributes from an obj.

    If d empty, use standard 'getattr'.
    """
    if len(d):
        v = d.values()[0]
        if isinstance(v, int):
            return AttrFromTuple(d)
        elif isinstance(v, str):
            return AttrFromObject(d)
        raise ValueError('dictionary values must be int or str!')

    return getattr


class AttrFromTuple(object):

    def __init__(self, attrDict):
        self.attrDict = attrDict

    def __call__(self, obj, attr, default=None):
        'getattr from tuple obj'
        try:
            return obj[self.attrDict[attr]]
        except (IndexError, KeyError):
            if default is not None:
                return default
        raise AttributeError("object has no attribute '%s'" % attr)


class AttrFromObject(AttrFromTuple):

    def __call__(self, obj, attr, default=None):
        'getattr with attribute name aliases'
        try:
            return getattr(obj, self.attrDict[attr])
        except KeyError:
            try:
                return getattr(obj, attr)
            except KeyError:
                if default is not None:
                    return default
        raise AttributeError("object has no attribute '%s'" % attr)


def kwargs_filter(kwargs, allowed):
    'return dictionary of kwargs filtered by list allowed'
    d = {}
    for arg in allowed:
        try:
            d[arg] = kwargs[arg]
        except KeyError:
            pass
    return d


def split_kwargs(kwargs, *targets):
    '''split kwargs into n+1 dicts for n targets; each target must
    be a list of arguments for that target'''
    kwargs = kwargs.copy()
    out = []
    for args in targets:
        d = {}
        for arg in args:
            try:
                d[arg] = kwargs[arg]
                del kwargs[arg]
            except KeyError:
                pass
        out.append(d)
    out.append(kwargs)
    return out



from seqdb import SequenceDB, BasicSeqInfoDict
from annotation import AnnotationDB, TranslationAnnot, TranslationAnnotSlice
import classutil
import sequence
import UserDict


class SeqTranslator(sequence.SequenceBase):
    """Translator object for positive or minus strand of a sequence.
    Slicing returns TranslationAnnotSlice of the appropriate
    TranslationAnnot representing one of the six possible frames for
    this sequence."""

    def __init__(self, db, id, reversePath=None):
        self.id = id
        sequence.SequenceBase.__init__(self)
        if reversePath: # create top-level object for reverse strand
            self.orientation = -1
            self.start = -len(self)
            self.stop = 0
            self._reverse = reversePath
        if self.id not in self.db.seqDB:
            raise KeyError('sequence %s not in db %s' % (self.id, self.db))

    def __getitem__(self, k):
        """get TranslationAnnotSlice for coordinates given by slice k """
        start = k.start                 # deal with [:stop] slices
        if start is None:
            start = self.start
        stop = k.stop                   # deal with [start:] slices
        if stop is None:
            stop = self.stop

        annoID = self._get_anno_id(start)
        a = self.db.annodb[annoID] # get TranslationAnnot object
        s = a.sequence # corresponding nucleotide region

        return a[(start - s.start) / 3: (stop - s.start) / 3]

    def absolute_slice(self, start, stop):
        """get protein slice in absolute nucleotide coords;
        perform negation before slicing """
        if start<0:
            return (-self)[start:stop]
        else:
            return self[start:stop]

    def __len__(self):
        return self.db.seqInfoDict[self.id].length

    def __neg__(self):
        """get SeqTranslator for the opposite strand """
        try:
            return self._reverse
        except AttributeError:
            self._reverse = self.__class__(self.db, self.id,
                                           reversePath=self)
            return self._reverse

    def _get_anno_id(self, start):
        """get annotation ID for frame starting at start """
        if self.orientation > 0: # positive strand
            return '%s:%d' % (self.id, start % 3)
        else: # negative strand
            return '%s:-%d' % (self.id, (-start) % 3)

    def iter_frames(self):
        'iterate over the 6 possible frames, yielding TranslationAnnot'
        for frame in ('0', '1', '2', '-0', '-1', '-2'):
            yield self.db.annodb['%s:%s' % (self.id, frame)]

    def __repr__(self):
        return 'SeqTranslator(' + sequence.SequenceBase.__repr__(self) + ')'


class TranslationDB(SequenceDB):
    """Provides an automatic translation interface for a nucleotide sequence
    database: slicing of top-level sequence objects will return the
    corresponding TranslationAnnotSlice for that slice, i.e. the
    translated protein sequence, rather than the nucleotide sequence. """
    itemClass = SeqTranslator
    _seqtype = sequence.DNA_SEQTYPE

    def __init__(self, seqDB, **kwargs):
        self.seqDB = seqDB
        try:
            self.seqInfoDict = seqDB.seqInfoDict
        except AttributeError:
            self.seqInfoDict = BasicSeqInfoDict(seqDB)
        self.annodb = AnnotationDB(SixFrameInfo(seqDB), seqDB,
                                   itemClass=TranslationAnnot,
                                   itemSliceClass=TranslationAnnotSlice,
                                   sliceAttrDict=dict(id=0, start=1, stop=2),
                                   checkFirstID=False)
        SequenceDB.__init__(self, **kwargs)


class SixFrameInfo(object, UserDict.DictMixin):
    """Dictionary of slice info for all six frames of each seq in seqDB. """

    def __init__(self, seqDB):
        self.seqDB = seqDB

    def __getitem__(self, k):
        "convert ID of form seqID:frame into slice info tuple"
        i = k.rfind(':')
        if i < 0:
            raise KeyError('invalid TranslationInfo key: %s' % (k, ))
        seqID = k[:i]
        length = len(self.seqDB[seqID]) # sequence length
        frame = int(k[i+1:])
        if k[i+1] == '-': # negative frame -0, -1, or -2
            return (seqID, -(length - ((length + frame) % 3)), frame)
        else: # positive frame 0, 1 or 2
            return (seqID, frame, length - ((length - frame) % 3))

    def __len__(self):
        return 6 * len(self.seqDB)

    def __iter__(self):
        for seqID in self.seqDB:
            for frame in (':0', ':1', ':2', ':-0', ':-1', ':-2'):
                yield seqID + frame

    def keys(self):
        return list(self)

    # these methods should not be implemented for read-only database.
    clear = setdefault = pop = popitem = copy = update = \
            classutil.read_only_error


def get_translation_db(seqDB):
    """Use cached seqDB.translationDB if already present, or create it """
    try:
        return seqDB.translationDB
    except AttributeError: # create a new TranslationAnnot DB
        tdb = TranslationDB(seqDB)
        try:
            seqDB.translationDB = tdb
        except AttributeError:
            pass # won't let us cache? Just hand back the TranslationDB
        return tdb



from pygr.cnestedlist import *
from specialseq import *


def maf2nclist(maffiles, stem):
    align = NLMSALetters(stem, 'w')
    seqs = {}
    for i in maffiles:
        f = open(i, 'r')
        if f.readline().split()[0] != '##maf':
            raise "Error processing %s: Invalid file format" % (i)

        l = f.readline()
        while l:
##            print l
            la = l.split()
            if(len(la) == 0 or la[0] == '#'):
                pass
            elif(la[0] == 'a'):
                storeMAFrecord(align, seqs, f)
            else:
                return
            l=f.readline()
    align.build()
    return align


def storeMAFrecord(align, seqs, fh):
    s = fh.readline().split()
    begin = align.seqlist[0].length
    while len(s) == 7 and s[0] == 's':
##        print s
        lpoStart = begin
        seqStart = int(s[2])
        try:
            seq = seqs[s[1]]
        except:
            seq = refSequence(s[1])
            seqs[s[1]] = seq
        if (s[4] == '-'):
            rev = True
        else:
            rev = False
        for ival in s[6].split('-'):
            if len(ival) > 0:
                if (rev):
                    align[-(lpoStart + len(ival)):-lpoStart] = \
                            seq[seqStart:seqStart + len(ival)]
                else:
                    align[lpoStart:lpoStart + len(ival)] = \
                            seq[seqStart:seqStart + len(ival)]
            seqStart += len(ival)
            lpoStart += len(ival) + 1
        s = fh.readline().split()


maf2nclist(['chrX.maf', ], 'testdb/chrX')



"""
Implements logging functionality

Upon import creates a module level log class (log) and
the following logging functions:

debug, info, warn and error

The default formatters will print out the function the log was triggered from.
"""

import logging
import sys

# python 2.5 the watershed release that introduced most changes since 2.1

PYTHON_25 = sys.version_info >= (2, 5)


def get_logger(name='pygr-log', stream=sys.stdout, formatter=None):
    """
    Returns a logger

    >>> disable('INFO')
    >>> info('logtest, this message SHOULD NOT be visible')
    >>> disable()
    >>> info('logtest, this message should be visible')
    >>> disable('DEBUG')
    >>> debug('logtest, this message SHOULD NOT be visible')
    >>> info('logtest, this message should be visible')
    """
    logger = logging.getLogger(name)

    # this is needed in case the process is
    # forked/multithreaded; loggers exist in a global scope
    # we don't want each import to duplocate this handler

    if not logger.handlers:
        console = logging.StreamHandler(stream)
        console.setLevel(logging.DEBUG)
        if PYTHON_25:
            format = '%(levelname)s %(module)s.%(funcName)s: %(message)s'
        else:
            format = '%(levelname)s %(module)s: %(message)s'

        formatter = formatter or logging.Formatter(format)
        console.setFormatter(formatter)
        logger.addHandler(console)
        logger.setLevel(logging.DEBUG)
    return logger


def disable(level=0):
    """
    Disables logging levels
    Levels: DEBUG, INFO, WARNING, ERROR

    >>> disable('INFO')
    >>> info('logtest, this message SHOULD NOT be visible')
    """
    level = str(level)
    value = dict(NOTSET=0, DEBUG=10, INFO=20, WARNING=30, ERROR=40)\
            .get(level.upper(), 0)
    logging.disable(value)


# populate some loggers by default
log = get_logger()
debug, info, warn, error = log.debug, log.info, log.warn, log.error


def test(verbose=0):
    "Performs module level testing"
    import doctest
    doctest.testmod(verbose=verbose)


if __name__ == "__main__":
    test()





from __future__ import generators
from mapping import *
from sequence import SequenceBase, DNA_SEQTYPE, RNA_SEQTYPE, PROTEIN_SEQTYPE
import types
from classutil import methodFactory, standard_getstate,\
     override_rich_cmp, generate_items, get_bound_subclass, standard_setstate,\
     get_valid_path, standard_invert, RecentValueDictionary, read_only_error,\
     SourceFileName, split_kwargs
import os
import platform
import UserDict
import warnings
import logger


class TupleDescriptor(object):
    'return tuple entry corresponding to named attribute'

    def __init__(self, db, attr):
        self.icol = db.data[attr] # index of this attribute in the tuple

    def __get__(self, obj, klass):
        return obj._data[self.icol]

    def __set__(self, obj, val):
        raise AttributeError('this database is read-only!')


class TupleIDDescriptor(TupleDescriptor):

    def __set__(self, obj, val):
        raise AttributeError('''You cannot change obj.id directly.
        Instead, use db[newID] = obj''')


class TupleDescriptorRW(TupleDescriptor):
    'read-write interface to named attribute'

    def __init__(self, db, attr):
        self.attr = attr
        self.icol = db.data[attr] # index of this attribute in the tuple
        self.attrSQL = db._attrSQL(attr, sqlColumn=True) # SQL column name

    def __set__(self, obj, val):
        obj.db._update(obj.id, self.attrSQL, val) # AND UPDATE THE DATABASE
        obj.save_local(self.attr, val)


class SQLDescriptor(object):
    'return attribute value by querying the database'

    def __init__(self, db, attr):
        self.selectSQL = db._attrSQL(attr) # SQL expression for this attr

    def __get__(self, obj, klass):
        return obj._select(self.selectSQL)

    def __set__(self, obj, val):
        raise AttributeError('this database is read-only!')


class SQLDescriptorRW(SQLDescriptor):
    'writeable proxy to corresponding column in the database'

    def __set__(self, obj, val):
        obj.db._update(obj.id, self.selectSQL, val) #just update the database


class ReadOnlyDescriptor(object):
    'enforce read-only policy, e.g. for ID attribute'

    def __init__(self, db, attr):
        self.attr = '_'+attr

    def __get__(self, obj, klass):
        return getattr(obj, self.attr)

    def __set__(self, obj, val):
        raise AttributeError('attribute %s is read-only!' % self.attr)


def select_from_row(row, what):
    "return value from SQL expression applied to this row"
    sql, params = row.db._format_query('select %s from %s where %s=%%s limit 2'
                                       % (what, row.db.name,
                                          row.db.primary_key), (row.id, ))
    row.db.cursor.execute(sql, params)
    t = row.db.cursor.fetchmany(2) # get at most two rows
    if len(t) != 1:
        raise KeyError('%s[%s].%s not found, or not unique'
                       % (row.db.name, str(row.id), what))
    return t[0][0] #return the single field we requested


def init_row_subclass(cls, db):
    'add descriptors for db attributes'
    for attr in db.data: # bind all database columns
        if attr == 'id': # handle ID attribute specially
            setattr(cls, attr, cls._idDescriptor(db, attr))
            continue
        try: # check if this attr maps to an SQL column
            db._attrSQL(attr, columnNumber=True)
        except AttributeError: # treat as SQL expression
            setattr(cls, attr, cls._sqlDescriptor(db, attr))
        else: # treat as interface to our stored tuple
            setattr(cls, attr, cls._columnDescriptor(db, attr))


def dir_row(self):
    """get list of column names as our attributes """
    return self.db.data.keys()


class TupleO(object):
    """Provides attribute interface to a database tuple.
    Storing the data as a tuple instead of a standard Python object
    (which is stored using __dict__) uses about five-fold less
    memory and is also much faster (the tuples returned from the
    DB API fetch are simply referenced by the TupleO, with no
    need to copy their individual values into __dict__).

    This class follows the 'subclass binding' pattern, which
    means that instead of using __getattr__ to process all
    attribute requests (which is un-modular and leads to all
    sorts of trouble), we follow Python's new model for
    customizing attribute access, namely Descriptors.
    We use classutil.get_bound_subclass() to automatically
    create a subclass of this class, calling its _init_subclass()
    class method to add all the descriptors needed for the
    database table to which it is bound.

    See the Pygr Developer Guide section of the docs for a
    complete discussion of the subclass binding pattern."""
    _columnDescriptor = TupleDescriptor
    _idDescriptor = TupleIDDescriptor
    _sqlDescriptor = SQLDescriptor
    _init_subclass = classmethod(init_row_subclass)
    _select = select_from_row
    __dir__ = dir_row

    def __init__(self, data):
        self._data = data # save our data tuple


def insert_and_cache_id(self, l, **kwargs):
    'insert tuple into db and cache its rowID on self'
    self.db._insert(l) # save to database
    try:
        rowID = kwargs['id']  # use the ID supplied by user
    except KeyError:
        rowID = self.db.get_insert_id() # get auto-inc ID value
    self.cache_id(rowID) # cache this ID on self


class TupleORW(TupleO):
    'read-write version of TupleO'
    _columnDescriptor = TupleDescriptorRW
    insert_and_cache_id = insert_and_cache_id

    def __init__(self, data, newRow=False, **kwargs):
        if not newRow: # just cache data from the database
            self._data = data
            return
        self._data = self.db.tuple_from_dict(kwargs) # convert to tuple
        self.insert_and_cache_id(self._data, **kwargs)

    def cache_id(self, row_id):
        self.save_local('id', row_id)

    def save_local(self, attr, val):
        icol = self._attrcol[attr]
        try:
            self._data[icol] = val # FINALLY UPDATE OUR LOCAL CACHE
        except TypeError: # TUPLE CAN'T STORE NEW VALUE, SO USE A LIST
            self._data = list(self._data)
            self._data[icol] = val # FINALLY UPDATE OUR LOCAL CACHE


TupleO._RWClass = TupleORW # record this as writeable interface class


class ColumnDescriptor(object):
    'read-write interface to column in a database, cached in obj.__dict__'

    def __init__(self, db, attr, readOnly = False):
        self.attr = attr
        # Map attr to SQL column name.
        self.col = db._attrSQL(attr, sqlColumn=True)
        self.db = db
        if readOnly:
            self.__class__ = self._readOnlyClass

    def __get__(self, obj, objtype):
        try:
            return obj.__dict__[self.attr]
        except KeyError: # NOT IN CACHE.  TRY TO GET IT FROM DATABASE
            if self.col==self.db.primary_key:
                raise AttributeError
            self.db._select('where %s=%%s' % self.db.primary_key, (obj.id, ),
                            self.col)
            l = self.db.cursor.fetchall()
            if len(l)!=1:
                raise AttributeError('db row not found or not unique!')
            obj.__dict__[self.attr] = l[0][0] # UPDATE THE CACHE
            return l[0][0]

    def __set__(self, obj, val):
        if not hasattr(obj, '_localOnly'): # ONLY CACHE, DON'T SAVE TO DATABASE
            self.db._update(obj.id, self.col, val) # UPDATE THE DATABASE
        obj.__dict__[self.attr] = val # UPDATE THE CACHE
##         try:
##             m = self.consequences
##         except AttributeError:
##             return
##         m(obj, val) # GENERATE CONSEQUENCES
##     def bind_consequences(self, f):
##         'make function f be run as consequences method whenever value is set'
##         import new
##         self.consequences = new.instancemethod(f, self, self.__class__)


class ReadOnlyColumnDesc(ColumnDescriptor):

    def __set__(self, obj, val):
        raise AttributeError('The ID of a database object is not writeable.')
ColumnDescriptor._readOnlyClass = ReadOnlyColumnDesc


class SQLRow(object):
    """Provide transparent interface to a row in the database: attribute access
       will be mapped to SELECT of the appropriate column, but data is not
       cached on this object.
    """
    _columnDescriptor = _sqlDescriptor = SQLDescriptor
    _idDescriptor = ReadOnlyDescriptor
    _init_subclass = classmethod(init_row_subclass)
    _select = select_from_row
    __dir__ = dir_row

    def __init__(self, rowID):
        self._id = rowID


class SQLRowRW(SQLRow):
    'read-write version of SQLRow'
    _columnDescriptor = SQLDescriptorRW
    insert_and_cache_id = insert_and_cache_id

    def __init__(self, rowID, newRow=False, **kwargs):
        if not newRow: # just cache data from the database
            return self.cache_id(rowID)
        l = self.db.tuple_from_dict(kwargs) # convert to tuple
        self.insert_and_cache_id(l, **kwargs)

    def cache_id(self, rowID):
        self._id = rowID


SQLRow._RWClass = SQLRowRW


def list_to_dict(names, values):
    'return dictionary of those named args that are present in values[]'
    d = {}
    for i, v in enumerate(values):
        try:
            d[names[i]] = v
        except IndexError:
            break
    return d


def get_name_cursor(name=None, **kwargs):
    '''get table name and cursor by parsing name or using configFile.
    If neither provided, will try to get via your MySQL config file.
    If connect is None, will use MySQLdb.connect()'''
    if name is not None:
        argList = name.split() # TREAT AS WS-SEPARATED LIST
        if len(argList) > 1:
            name = argList[0] # USE 1ST ARG AS TABLE NAME
            argnames = ('host', 'user', 'passwd') # READ ARGS IN THIS ORDER
            kwargs = kwargs.copy() # a copy we can overwrite
            kwargs.update(list_to_dict(argnames, argList[1:]))
    serverInfo = DBServerInfo(**kwargs)
    return name, serverInfo.cursor(), serverInfo


def mysql_connect(connect=None, configFile=None, useStreaming=False, **args):
    """return connection and cursor objects, using .my.cnf if necessary"""
    kwargs = args.copy() # a copy we can modify
    if 'user' not in kwargs and configFile is None: #Find where config file is
        osname = platform.system()
        if osname in('Microsoft', 'Windows'): # Machine is a Windows box
            paths = []
            try: # handle case where WINDIR not defined by Windows...
                windir = os.environ['WINDIR']
                paths += [(windir, 'my.ini'), (windir, 'my.cnf')]
            except KeyError:
                pass
            try:
                sysdrv = os.environ['SYSTEMDRIVE']
                paths += [(sysdrv, os.path.sep + 'my.ini'),
                          (sysdrv, os.path.sep + 'my.cnf')]
            except KeyError:
                pass
            if len(paths) > 0:
                configFile = get_valid_path(*paths)
        else: # treat as normal platform with home directories
            configFile = os.path.join(os.path.expanduser('~'), '.my.cnf')

    # allows for a local mysql local configuration file to be read
    # from the current directory
    configFile = configFile or os.path.join(os.getcwd(), 'mysql.cnf')

    if configFile and os.path.exists(configFile):
        kwargs['read_default_file'] = configFile
        connect = None # force it to use MySQLdb
    if connect is None:
        import MySQLdb
        connect = MySQLdb.connect
        kwargs['compress'] = True
    if useStreaming:  # use server side cursors for scalable result sets
        try:
            from MySQLdb import cursors
            kwargs['cursorclass'] = cursors.SSCursor
        except (ImportError, AttributeError):
            pass
    conn = connect(**kwargs)
    cursor = conn.cursor()
    return conn, cursor


_mysqlMacros = dict(IGNORE='ignore', REPLACE='replace',
                    AUTO_INCREMENT='AUTO_INCREMENT', SUBSTRING='substring',
                    SUBSTR_FROM='FROM', SUBSTR_FOR='FOR')


def mysql_table_schema(self, analyzeSchema=True):
    'retrieve table schema from a MySQL database, save on self'
    import MySQLdb
    self._format_query = SQLFormatDict(MySQLdb.paramstyle, _mysqlMacros)
    if not analyzeSchema:
        return
    self.clear_schema() # reset settings and dictionaries
    self.cursor.execute('describe %s' % self.name) # get info about columns
    columns = self.cursor.fetchall()
    self.cursor.execute('select * from %s limit 1' % self.name) # descriptions
    for icol, c in enumerate(columns):
        field = c[0]
        self.columnName.append(field) # list of columns in same order as table
        if c[3] == "PRI": # record as primary key
            if self.primary_key is None:
                self.primary_key = field
            else:
                try:
                    self.primary_key.append(field)
                except AttributeError:
                    self.primary_key = [self.primary_key, field]
            if c[1][:3].lower() == 'int':
                self.usesIntID = True
            else:
                self.usesIntID = False
        elif c[3] == "MUL":
            self.indexed[field] = icol
        self.description[field] = self.cursor.description[icol]
        self.columnType[field] = c[1] # SQL COLUMN TYPE


_sqliteMacros = dict(IGNORE='or ignore', REPLACE='insert or replace',
                     AUTO_INCREMENT='', SUBSTRING='substr',
                    SUBSTR_FROM=',', SUBSTR_FOR=',')


def import_sqlite():
    'import sqlite3 (for Python 2.5+) or pysqlite2 for earlier Python versions'
    try:
        import sqlite3 as sqlite
    except ImportError:
        from pysqlite2 import dbapi2 as sqlite
    return sqlite


def sqlite_table_schema(self, analyzeSchema=True):
    'retrieve table schema from a sqlite3 database, save on self'
    sqlite = import_sqlite()
    self._format_query = SQLFormatDict(sqlite.paramstyle, _sqliteMacros)
    if not analyzeSchema:
        return
    self.clear_schema() # reset settings and dictionaries
    self.cursor.execute('PRAGMA table_info("%s")' % self.name)
    columns = self.cursor.fetchall()
    self.cursor.execute('select * from %s limit 1' % self.name) # descriptions
    for icol, c in enumerate(columns):
        field = c[1]
        self.columnName.append(field) # list of columns in same order as table
        self.description[field] = self.cursor.description[icol]
        self.columnType[field] = c[2] # SQL COLUMN TYPE
    # Get primary key / unique indexes.
    self.cursor.execute('select name from sqlite_master where tbl_name="%s" \
                        and type="index" and sql is null' % self.name)
    for indexname in self.cursor.fetchall(): # search indexes for primary key
        self.cursor.execute('PRAGMA index_info("%s")' % indexname)
        l = self.cursor.fetchall() # get list of columns in this index
        if len(l) == 1: # assume 1st single-column unique index is primary key!
            self.primary_key = l[0][2]
            break # done searching for primary key!
    if self.primary_key is None:
        # Grrr, INTEGER PRIMARY KEY handled differently.
        self.cursor.execute('select sql from sqlite_master where \
                            tbl_name="%s" and type="table"' % self.name)
        sql = self.cursor.fetchall()[0][0]
        for columnSQL in sql[sql.index('(') + 1:].split(','):
            if 'primary key' in columnSQL.lower(): # must be the primary key!
                col = columnSQL.split()[0] # get column name
                if col in self.columnType:
                    self.primary_key = col
                    break # done searching for primary key!
                else:
                    raise ValueError('unknown primary key %s in table %s'
                                     % (col, self.name))
    if self.primary_key is not None: # check its type
        if self.columnType[self.primary_key] == 'int' or \
               self.columnType[self.primary_key] == 'integer':
            self.usesIntID = True
        else:
            self.usesIntID = False


class SQLFormatDict(object):
    '''Perform SQL keyword replacements for maintaining compatibility across
    a wide range of SQL backends.  Uses Python dict-based string format
    function to do simple string replacements, and also to convert
    params list to the paramstyle required for this interface.
    Create by passing a dict of macros and the db-api paramstyle:
    sfd = SQLFormatDict("qmark", substitutionDict)

    Then transform queries+params as follows; input should be "format" style:
    sql,params = sfd("select * from foo where id=%s and val=%s", (myID,myVal))
    cursor.execute(sql, params)
    '''
    _paramFormats = dict(pyformat='%%(%d)s', numeric=':%d', named=':%d',
                         qmark='(ignore)', format='(ignore)')

    def __init__(self, paramstyle, substitutionDict={}):
        self.substitutionDict = substitutionDict.copy()
        self.paramstyle = paramstyle
        self.paramFormat = self._paramFormats[paramstyle]
        self.makeDict = (paramstyle == 'pyformat' or paramstyle == 'named')
        if paramstyle == 'qmark': # handle these as simple substitution
            self.substitutionDict['?'] = '?'
        elif paramstyle == 'format':
            self.substitutionDict['?'] = '%s'

    def __getitem__(self, k):
        'apply correct substitution for this SQL interface'
        try:
            return self.substitutionDict[k] # apply our substitutions
        except KeyError:
            pass
        if k == '?': # sequential parameter
            s = self.paramFormat % self.iparam
            self.iparam += 1 # advance to the next parameter
            return s
        raise KeyError('unknown macro: %s' % k)

    def __call__(self, sql, paramList):
        'returns corrected sql,params for this interface'
        self.iparam = 1 # DB-ABI param indexing begins at 1
        sql = sql.replace('%s', '%(?)s') # convert format into pyformat
        s = sql % self # apply all %(x)s replacements in sql
        if self.makeDict: # construct a params dict
            paramDict = {}
            for i, param in enumerate(paramList):
                # i + 1 because DB-ABI parameter indexing begins at 1
                paramDict[str(i + 1)] = param
            return s, paramDict
        else: # just return the original params list
            return s, paramList


def get_table_schema(self, analyzeSchema=True):
    'run the right schema function based on type of db server connection'
    try:
        modname = self.cursor.__class__.__module__
    except AttributeError:
        raise ValueError('no cursor object or module information!')
    try:
        schema_func = self._schemaModuleDict[modname]
    except KeyError:
        raise KeyError('''unknown db module: %s. Use _schemaModuleDict
        attribute to supply a method for obtaining table schema
        for this module''' % modname)
    schema_func(self, analyzeSchema) # run the schema function


_schemaModuleDict = {'MySQLdb.cursors': mysql_table_schema,
                     'pysqlite2.dbapi2': sqlite_table_schema,
                     'sqlite3': sqlite_table_schema}


class SQLTableBase(object, UserDict.DictMixin):
    "Store information about an SQL table as dict keyed by primary key"
    _schemaModuleDict = _schemaModuleDict # default module list
    get_table_schema = get_table_schema

    def __init__(self, name, cursor=None, itemClass=None, attrAlias=None,
                 clusterKey=None, createTable=None, graph=None, maxCache=None,
                 arraysize=1024, itemSliceClass=None, dropIfExists=False,
                 serverInfo=None, autoGC=True, orderBy=None,
                 writeable=False, iterSQL=None, iterColumns=None,
                 primaryKey=None, **kwargs):
        if autoGC: # automatically garbage collect unused objects
            self._weakValueDict = RecentValueDictionary(autoGC) # object cache
        else:
            self._weakValueDict = {}
        self.autoGC = autoGC
        self.orderBy = orderBy
        if orderBy and serverInfo and serverInfo._serverType == 'mysql':
            if iterSQL and iterColumns: # both required for mysql!
                self.iterSQL, self.iterColumns = iterSQL, iterColumns
            else:
                raise ValueError('For MySQL tables with orderBy, you MUST \
                                 specify iterSQL and iterColumns as well!')

        self.writeable = writeable
        if cursor is None:
            if serverInfo is not None: # get cursor from serverInfo
                cursor = serverInfo.cursor()
            else: # try to read connection info from name or config file
                name, cursor, serverInfo = get_name_cursor(name, **kwargs)
        else:
            warnings.warn("The cursor argument is deprecated. Use serverInfo \
                          instead!", DeprecationWarning, stacklevel=2)
        self.cursor = cursor
        if createTable is not None: # RUN COMMAND TO CREATE THIS TABLE
            if dropIfExists: # get rid of any existing table
                cursor.execute('drop table if exists ' + name)
            self.get_table_schema(False) # check dbtype, init _format_query
            sql, params = self._format_query(createTable, ()) # apply macros
            cursor.execute(sql) # create the table
        self.name = name
        if graph is not None:
            self.graph = graph
        if maxCache is not None:
            self.maxCache = maxCache
        if arraysize is not None:
            self.arraysize = arraysize
            cursor.arraysize = arraysize
        self.get_table_schema() # get schema of columns to serve as attrs
        if primaryKey is not None:
            self.primary_key = primaryKey
            self.primaryKey = primaryKey
        self.data = {} # map of all attributes, including aliases
        for icol, field in enumerate(self.columnName):
            self.data[field] = icol # 1st add mappings to columns
        try:
            self.data['id'] = self.data[self.primary_key]
        except (KeyError, TypeError):
            pass
        if hasattr(self, '_attr_alias'):
            # Apply attribute aliases for this class.
            self.addAttrAlias(False, **self._attr_alias)
        self.objclass(itemClass) # NEED TO SUBCLASS OUR ITEM CLASS
        if itemSliceClass is not None:
            self.itemSliceClass = itemSliceClass
            # Need to subclass itemSliceClass.
            get_bound_subclass(self, 'itemSliceClass', self.name)
        if attrAlias is not None: # ADD ATTRIBUTE ALIASES
            self.attrAlias = attrAlias # RECORD FOR PICKLING PURPOSES
            self.data.update(attrAlias)
        if clusterKey is not None:
            self.clusterKey = clusterKey
        if serverInfo is not None:
            self.serverInfo = serverInfo

    def __len__(self):
        self._select(selectCols = 'count(*)')
        return self.cursor.fetchone()[0]

    def __hash__(self):
        return id(self)

    def __cmp__(self, other):
        'only match self and no other!'
        if self is other:
            return 0
        else:
            return cmp(id(self), id(other))
    _pickleAttrs = dict(name=0, clusterKey=0, maxCache=0, arraysize=0,
                        attrAlias=0, serverInfo=0, autoGC=0, orderBy=0,
                        writeable=0, iterSQL=0, iterColumns=0, primaryKey=0)
    __getstate__ = standard_getstate

    def __setstate__(self, state):
        # default cursor provisioning by worldbase is deprecated!
        ## if 'serverInfo' not in state: # hmm, no address for db server?
        ##     try: # SEE IF WE CAN GET CURSOR DIRECTLY FROM RESOURCE DATABASE
        ##         from Data import getResource
        ##         state['cursor'] = getResource.getTableCursor(state['name'])
        ##     except ImportError:
        ##         pass # FAILED, SO TRY TO GET A CURSOR IN THE USUAL WAYS...
        self.__init__(**state)

    def __repr__(self):
        return '<SQL table ' + self.name + '>'

    def clear_schema(self):
        'reset all schema information for this table'
        self.description={}
        self.columnName = []
        self.columnType = {}
        self.usesIntID = None
        self.primary_key = None
        self.indexed = {}

    def _attrSQL(self, attr, sqlColumn=False, columnNumber=False):
        "Translate python attribute name to appropriate SQL expression"
        try: # MAKE SURE THIS ATTRIBUTE CAN BE MAPPED TO DATABASE EXPRESSION
            field = self.data[attr]
        except KeyError:
            raise AttributeError('attribute %s not a valid column \
                                 or alias in %s' % (attr, self.name))
        if sqlColumn: # ENSURE THAT THIS TRULY MAPS TO A COLUMN NAME IN THE DB
            try: # CHECK IF field IS COLUMN NUMBER
                return self.columnName[field] # RETURN SQL COLUMN NAME
            except TypeError:
                try:
                    # Check if field is SQL column name, return it if so.
                    return self.columnName[self.data[field]]
                except (KeyError, TypeError):
                    raise AttributeError('attribute %s does not map to an SQL \
                                         column in %s' % (attr, self.name))
        if columnNumber:
            try: # CHECK IF field IS A COLUMN NUMBER
                return field + 0 # ONLY RETURN AN INTEGER
            except TypeError:
                try: # CHECK IF field IS ITSELF THE SQL COLUMN NAME
                    return self.data[field] + 0 # ONLY RETURN AN INTEGER
                except (KeyError, TypeError):
                    raise ValueError('attribute %s does not map to a SQL \
                                     column!' % attr)
        if isinstance(field, types.StringType):
            # Use aliased expression for database select instead of attr.
            attr = field
        elif attr == 'id':
            attr = self.primary_key
        return attr

    def addAttrAlias(self, saveToPickle=True, **kwargs):
        """Add new attributes as aliases of existing attributes.
           They can be specified either as named args:
           t.addAttrAlias(newattr=oldattr)
           or by passing a dictionary kwargs whose keys are newattr
           and values are oldattr:
           t.addAttrAlias(**kwargs)
           saveToPickle=True forces these aliases to be saved if object
           is pickled.
        """
        if saveToPickle:
            self.attrAlias.update(kwargs)
        for key, val in kwargs.items():
            try: # 1st CHECK WHETHER val IS AN EXISTING COLUMN / ALIAS
                self.data[val] + 0 # CHECK WHETHER val MAPS TO A COLUMN NUMBER
                # Yes, val is an actual SQL column name, so save it directly.
                raise KeyError
            except TypeError: # val IS ITSELF AN ALIAS
                self.data[key] = self.data[val] # SO MAP TO WHAT IT MAPS TO
            except KeyError: # TREAT AS ALIAS TO SQL EXPRESSION
                self.data[key] = val

    def objclass(self, oclass=None):
        """Create class representing a row in this table
        by subclassing oclass, adding data"""
        if oclass is not None: # use this as our base itemClass
            self.itemClass = oclass
        if self.writeable:
            # Use its writeable version.
            self.itemClass = self.itemClass._RWClass
        # Bind itemClass.
        oclass = get_bound_subclass(self, 'itemClass', self.name,
                                    subclassArgs=dict(db=self))
        if issubclass(oclass, TupleO):
            # Bind attribute list to tupleo interface.
            oclass._attrcol = self.data
        if hasattr(oclass, '_tableclass') and \
           not isinstance(self, oclass._tableclass):
            # Row class can override our current table class.
            self.__class__ = oclass._tableclass

    def _select(self, whereClause='', params=(), selectCols='t1.*',
                cursor=None, orderBy='', limit=''):
        'execute the specified query but do not fetch'
        sql, params = self._format_query('select %s from %s t1 %s %s %s'
                            % (selectCols, self.name, whereClause, orderBy,
                               limit), params)
        if cursor is None:
            self.cursor.execute(sql, params)
        else:
            cursor.execute(sql, params)

    def select(self, whereClause, params=None, oclass=None, selectCols='t1.*'):
        "Generate the list of objects that satisfy the database SELECT"
        if oclass is None:
            oclass = self.itemClass
        self._select(whereClause, params, selectCols)
        l = self.cursor.fetchall()
        for t in l:
            yield self.cacheItem(t, oclass)

    def query(self, **kwargs):
        'query for intersection of all specified kwargs, returned as iterator'
        criteria = []
        params = []
        for k, v in kwargs.items(): # CONSTRUCT THE LIST OF WHERE CLAUSES
            if v is None: # CONVERT TO SQL NULL TEST
                criteria.append('%s IS NULL' % self._attrSQL(k))
            else: # TEST FOR EQUALITY
                criteria.append('%s=%%s' % self._attrSQL(k))
                params.append(v)
        return self.select('where ' + ' and '.join(criteria), params)

    def _update(self, row_id, col, val):
        'update a single field in the specified row to the specified value'
        sql, params = self._format_query('update %s set %s=%%s where %s=%%s'
                                         % (self.name, col, self.primary_key),
                                         (val, row_id))
        self.cursor.execute(sql, params)

    def getID(self, t):
        try:
            return t[self.data['id']] # GET ID FROM TUPLE
        except TypeError: # treat as alias
            return t[self.data[self.data['id']]]

    def cacheItem(self, t, oclass):
        'get obj from cache if possible, or construct from tuple'
        try:
            id = self.getID(t)
        except KeyError: # NO PRIMARY KEY?  IGNORE THE CACHE.
            return oclass(t)
        try: # IF ALREADY LOADED IN OUR DICTIONARY, JUST RETURN THAT ENTRY
            return self._weakValueDict[id]
        except KeyError:
            pass
        o = oclass(t)
        self._weakValueDict[id] = o   # CACHE THIS ITEM IN OUR DICTIONARY
        return o

    def cache_items(self, rows, oclass=None):
        if oclass is None:
            oclass = self.itemClass
        for t in rows:
            yield self.cacheItem(t, oclass)

    def foreignKey(self, attr, k):
        'get iterator for objects with specified foreign key value'
        return self.select('where %s=%%s' % attr, (k, ))

    def limit_cache(self):
        'APPLY maxCache LIMIT TO CACHE SIZE'
        try:
            if self.maxCache<len(self._weakValueDict):
                self._weakValueDict.clear()
        except AttributeError:
            pass

    def get_new_cursor(self):
        """Return a new cursor object, or None if not possible """
        try:
            new_cursor = self.serverInfo.new_cursor
        except AttributeError:
            return None
        return new_cursor(self.arraysize)

    def generic_iterator(self, cursor=None, fetch_f=None, cache_f=None,
                         map_f=iter, cursorHolder=None):
        """generic iterator that runs fetch, cache and map functions.
        cursorHolder is used only to keep a ref in this function's locals,
        so that if it is prematurely terminated (by deleting its
        iterator), cursorHolder.__del__() will close the cursor."""
        if fetch_f is None: # JUST USE CURSOR'S PREFERRED CHUNK SIZE
            if cursor is None:
                fetch_f = self.cursor.fetchmany
            else:  # isolate this iter from other queries
                fetch_f = cursor.fetchmany
        if cache_f is None:
            cache_f = self.cache_items
        while True:
            self.limit_cache()
            rows = fetch_f() # FETCH THE NEXT SET OF ROWS
            if len(rows) == 0: # NO MORE DATA SO ALL DONE
                break
            for v in map_f(cache_f(rows)): # CACHE AND GENERATE RESULTS
                yield v

    def tuple_from_dict(self, d):
        'transform kwarg dict into tuple for storing in database'
        l = [None] * len(self.description) # DEFAULT COLUMN VALUES ARE NULL
        for col, icol in self.data.items():
            try:
                l[icol] = d[col]
            except (KeyError, TypeError):
                pass
        return l

    def tuple_from_obj(self, obj):
        'transform object attributes into tuple for storing in database'
        l = [None] * len(self.description) # DEFAULT COLUMN VALUES ARE NULL
        for col, icol in self.data.items():
            try:
                l[icol] = getattr(obj, col)
            except (AttributeError, TypeError):
                pass
        return l

    def _insert(self, l):
        '''insert tuple into the database.  Note this uses the MySQL
        extension REPLACE, which overwrites any duplicate key.'''
        s = '%(REPLACE)s into ' + self.name + ' values (' \
            + ','.join(['%s']*len(l)) + ')'
        sql, params = self._format_query(s, l)
        self.cursor.execute(sql, params)

    def insert(self, obj):
        '''insert new row by transforming obj to tuple of values'''
        l = self.tuple_from_obj(obj)
        self._insert(l)

    def get_insert_id(self):
        'get the primary key value for the last INSERT'
        try: # ATTEMPT TO GET ASSIGNED ID FROM DB
            auto_id = self.cursor.lastrowid
        except AttributeError: # CURSOR DOESN'T SUPPORT lastrowid
            raise NotImplementedError('''your db lacks lastrowid support?''')
        if auto_id is None:
            raise ValueError('lastrowid is None so cannot get ID from INSERT!')
        return auto_id

    def new(self, **kwargs):
        'return a new record with the assigned attributes, added to DB'
        if not self.writeable:
            raise ValueError('this database is read only!')
        obj = self.itemClass(None, newRow=True, **kwargs) # saves itself to db
        self._weakValueDict[obj.id] = obj # AND SAVE TO OUR LOCAL DICT CACHE
        return obj

    def clear_cache(self):
        'empty the cache'
        self._weakValueDict.clear()

    def __delitem__(self, k):
        if not self.writeable:
            raise ValueError('this database is read only!')
        sql, params = self._format_query('delete from %s where %s=%%s'
                                         % (self.name, self.primary_key),
                                         (k, ))
        self.cursor.execute(sql, params)
        try:
            del self._weakValueDict[k]
        except KeyError:
            pass


def getKeys(self, queryOption='', selectCols=None):
    'uses db select; does not force load'
    if selectCols is None:
        selectCols=self.primary_key
    if queryOption=='' and self.orderBy is not None:
        queryOption = self.orderBy # apply default ordering
    self.cursor.execute('select %s from %s %s'
                        % (selectCols, self.name, queryOption))
    # Get all at once, since other calls may reuse this cursor.
    return [t[0] for t in self.cursor.fetchall()]


def iter_keys(self, selectCols=None, orderBy='', map_f=iter,
              cache_f=lambda x: [t[0] for t in x], get_f=None, **kwargs):
    'guarantee correct iteration insulated from other queries'
    if selectCols is None:
        selectCols = self.primary_key
    if orderBy == '' and self.orderBy is not None:
        orderBy = self.orderBy # apply default ordering
    cursor = self.get_new_cursor()
    if cursor: # got our own cursor, guaranteeing query isolation
        if hasattr(self.serverInfo, 'iter_keys') \
           and self.serverInfo.custom_iter_keys:
            # use custom iter_keys() method from serverInfo
            return self.serverInfo.iter_keys(self, cursor,
                                             selectCols=selectCols,
                                             map_f=map_f, orderBy=orderBy,
                                             cache_f=cache_f, **kwargs)
        else:
            self._select(cursor=cursor, selectCols=selectCols,
                         orderBy=orderBy, **kwargs)
            return self.generic_iterator(cursor=cursor, cache_f=cache_f,
                                         map_f=map_f,
                                         cursorHolder=CursorCloser(cursor))
    else: # must pre-fetch all keys to ensure query isolation
        if get_f is not None:
            return iter(get_f())
        else:
            return iter(self.keys())


class SQLTable(SQLTableBase):
    """Provide on-the-fly access to rows in the database, caching
    the results in dict"""
    itemClass = TupleO # our default itemClass; constructor can override
    keys = getKeys
    __iter__ = iter_keys

    def load(self, oclass=None):
        "Load all data from the table"
        try: # IF ALREADY LOADED, NO NEED TO DO ANYTHING
            return self._isLoaded
        except AttributeError:
            pass
        if oclass is None:
            oclass = self.itemClass
        self.cursor.execute('select * from %s' % self.name)
        l = self.cursor.fetchall()
        self._weakValueDict = {} # just store the whole dataset in memory
        for t in l:
            self.cacheItem(t, oclass) # CACHE IT IN LOCAL DICTIONARY
        self._isLoaded = True # MARK THIS CONTAINER AS FULLY LOADED

    def __getitem__(self, k): # FIRST TRY LOCAL INDEX, THEN TRY DATABASE
        try:
            return self._weakValueDict[k] # DIRECTLY RETURN CACHED VALUE
        except KeyError: # NOT FOUND, SO TRY THE DATABASE
            sql, params = self._format_query('select * from %s where %s=%%s \
                                             limit 2' % (self.name,
                                                         self.primary_key),
                                             (k, ))
            self.cursor.execute(sql, params)
            l = self.cursor.fetchmany(2) # get at most 2 rows
            if len(l) != 1:
                raise KeyError('%s not found in %s, or not unique'
                               % (str(k), self.name))
            self.limit_cache()
            # Cache it in local dictionary.
            return self.cacheItem(l[0], self.itemClass)

    def __setitem__(self, k, v):
        if not self.writeable:
            raise ValueError('this database is read only!')
        try:
            if v.db is not self:
                raise AttributeError
        except AttributeError:
            raise ValueError('object not bound to itemClass for this db!')
        try:
            oldID = v.id
            if oldID is None:
                raise AttributeError
        except AttributeError:
            pass
        else: # delete row with old ID
            del self[v.id]
        v.cache_id(k) # cache the new ID on the object
        self.insert(v) # SAVE TO THE RELATIONAL DB SERVER
        self._weakValueDict[k] = v   # CACHE THIS ITEM IN OUR DICTIONARY

    def items(self):
        'forces load of entire table into memory'
        self.load()
        return [(k, self[k]) for k in self] # apply orderBy rules...

    def iteritems(self):
        'uses arraysize / maxCache and fetchmany() to manage data transfer'
        return iter_keys(self, selectCols='*', cache_f=None,
                         map_f=generate_items, get_f=self.items)

    def values(self):
        'forces load of entire table into memory'
        self.load()
        return [self[k] for k in self] # apply orderBy rules...

    def itervalues(self):
        'uses arraysize / maxCache and fetchmany() to manage data transfer'
        return iter_keys(self, selectCols='*', cache_f=None, get_f=self.values)


def getClusterKeys(self, queryOption=''):
    'uses db select; does not force load'
    self.cursor.execute('select distinct %s from %s %s'
                        % (self.clusterKey, self.name, queryOption))
    # Get all at once, since other calls may reuse this cursor.
    return [t[0] for t in self.cursor.fetchall()]


class SQLTableClustered(SQLTable):
    '''use clusterKey to load a whole cluster of rows at once,
       specifically, all rows that share the same clusterKey value.'''

    def __init__(self, *args, **kwargs):
        kwargs = kwargs.copy() # get a copy we can alter
        kwargs['autoGC'] = False # don't use WeakValueDictionary
        SQLTable.__init__(self, *args, **kwargs)
        if not self.orderBy: # add default ordering by clusterKey
            self.orderBy = 'ORDER BY %s,%s' % (self.clusterKey,
                                               self.primary_key)
            self.iterColumns = (self.clusterKey, self.clusterKey,
                                self.primary_key)
            self.iterSQL = 'WHERE %s>%%s or (%s=%%s and %s>%%s)' \
                           % self.iterColumns

    def clusterkeys(self):
        return getClusterKeys(self, 'order by %s' % self.clusterKey)

    def __getitem__(self, k):
        try:
            return self._weakValueDict[k] # DIRECTLY RETURN CACHED VALUE
        except KeyError: # NOT FOUND, SO TRY THE DATABASE
            sql, params = self._format_query('select t2.* from %s t1,%s t2 \
                                             where t1.%s=%%s and t1.%s=t2.%s'
                                             % (self.name, self.name,
                                                self.primary_key,
                                                self.clusterKey,
                                                self.clusterKey), (k, ))
            self.cursor.execute(sql, params)
            l = self.cursor.fetchall()
            self.limit_cache()
            for t in l: # LOAD THE ENTIRE CLUSTER INTO OUR LOCAL CACHE
                self.cacheItem(t, self.itemClass)
            return self._weakValueDict[k] # should be in cache, if row k exists

    def itercluster(self, cluster_id):
        'iterate over all items from the specified cluster'
        self.limit_cache()
        return self.select('where %s=%%s' % self.clusterKey, (cluster_id, ))


class SQLForeignRelation(object):
    'mapping based on matching a foreign key in an SQL table'

    def __init__(self, table, keyName):
        self.table = table
        self.keyName = keyName

    def __getitem__(self, k):
        'get list of objects o with getattr(o,keyName)==k.id'
        l = []
        for o in self.table.select('where %s=%%s' % self.keyName, (k.id, )):
            l.append(o)
        if len(l) == 0:
            raise KeyError('%s not found in %s' % (str(k), self.name))
        return l


class SQLTableNoCache(SQLTableBase):
    '''Provide on-the-fly access to rows in the database;
    values are simply an object interface (SQLRow) to back-end db query.
    Row data are not stored locally, but always accessed by querying the db'''
    itemClass = SQLRow # DEFAULT OBJECT CLASS FOR ROWS...
    keys = getKeys
    __iter__ = iter_keys

    def getID(self, t):
        return t[0] # GET ID FROM TUPLE

    def select(self, whereClause, params):
        return SQLTableBase.select(self, whereClause, params, self.oclass,
                                   self._attrSQL('id'))

    def __getitem__(self, k): # FIRST TRY LOCAL INDEX, THEN TRY DATABASE
        try:
            return self._weakValueDict[k] # DIRECTLY RETURN CACHED VALUE
        except KeyError: # NOT FOUND, SO TRY THE DATABASE
            self._select('where %s=%%s' % self.primary_key, (k, ),
                         self.primary_key)
            t = self.cursor.fetchmany(2)
            if len(t) != 1:
                raise KeyError('id %s non-existent or not unique' % k)
            o = self.itemClass(k) # create obj referencing this ID
            self._weakValueDict[k] = o # cache the SQLRow object
            return o

    def __setitem__(self, k, v):
        if not self.writeable:
            raise ValueError('this database is read only!')
        try:
            if v.db is not self:
                raise AttributeError
        except AttributeError:
            raise ValueError('object not bound to itemClass for this db!')
        try:
            del self[k] # delete row with new ID if any
        except KeyError:
            pass
        try:
            del self._weakValueDict[v.id] # delete from old cache location
        except KeyError:
            pass
        self._update(v.id, self.primary_key, k) # just change its ID in db
        v.cache_id(k) # change the cached ID value
        self._weakValueDict[k] = v # assign to new cache location

    def addAttrAlias(self, **kwargs):
        self.data.update(kwargs) # ALIAS KEYS TO EXPRESSION VALUES


# SQLRow is for non-caching table interface.
SQLRow._tableclass = SQLTableNoCache


class SQLTableMultiNoCache(SQLTableBase):
    "Trivial on-the-fly access for table with key that returns multiple rows"
    itemClass = TupleO # default itemClass; constructor can override
    _distinct_key = 'id' # DEFAULT COLUMN TO USE AS KEY

    def __init__(self, *args, **kwargs):
        SQLTableBase.__init__(self, *args, **kwargs)
        self.distinct_key = self._attrSQL(self._distinct_key)
        if not self.orderBy:
            self.orderBy = 'GROUP BY %s ORDER BY %s' % (self.distinct_key,
                                                        self.distinct_key)
            self.iterSQL = 'WHERE %s>%%s' % self.distinct_key
            self.iterColumns = (self.distinct_key, )

    def keys(self):
        return getKeys(self, selectCols=self.distinct_key)

    def __iter__(self):
        return iter_keys(self, selectCols=self.distinct_key)

    def __getitem__(self, id):
        sql, params = self._format_query('select * from %s where %s=%%s'
                                         % (self.name,
                                            self._attrSQL(self._distinct_key)),
                                         (id, ))
        self.cursor.execute(sql, params)
        # Prefetch all rows, since cursor may be reused.
        l = self.cursor.fetchall()
        for row in l:
            yield self.itemClass(row)

    def addAttrAlias(self, **kwargs):
        self.data.update(kwargs) # ALIAS KEYS TO EXPRESSION VALUES


class SQLEdges(SQLTableMultiNoCache):
    '''provide iterator over edges as (source, target, edge)
       and getitem[edge] --> [(source,target),...]'''
    _distinct_key = 'edge_id'
    _pickleAttrs = SQLTableMultiNoCache._pickleAttrs.copy()
    _pickleAttrs.update(dict(graph=0))

    def keys(self):
        self.cursor.execute('select %s,%s,%s from %s where %s is not null \
                            order by %s,%s' % (self._attrSQL('source_id'),
                                               self._attrSQL('target_id'),
                                               self._attrSQL('edge_id'),
                                               self.name,
                                               self._attrSQL('target_id'),
                                               self._attrSQL('source_id'),
                                               self._attrSQL('target_id')))
        l = [] # PREFETCH ALL ROWS, SINCE CURSOR MAY BE REUSED
        for source_id, target_id, edge_id in self.cursor.fetchall():
            l.append((self.graph.unpack_source(source_id),
                      self.graph.unpack_target(target_id),
                      self.graph.unpack_edge(edge_id)))
        return l

    __call__ = keys

    def __iter__(self):
        return iter(self.keys())

    def __getitem__(self, edge):
        sql, params = self._format_query('select %s,%s from %s where %s=%%s'
                                         % (self._attrSQL('source_id'),
                                            self._attrSQL('target_id'),
                                            self.name,
                                            self._attrSQL(self._distinct_key)),
                                         (self.graph.pack_edge(edge), ))
        self.cursor.execute(sql, params)
        l = [] # PREFETCH ALL ROWS, SINCE CURSOR MAY BE REUSED
        for source_id, target_id in self.cursor.fetchall():
            l.append((self.graph.unpack_source(source_id),
                      self.graph.unpack_target(target_id)))
        return l


class SQLEdgeDict(object):
    '2nd level graph interface to SQL database'

    def __init__(self, fromNode, table):
        self.fromNode = fromNode
        self.table = table
        if not hasattr(self.table, 'allowMissingNodes'):
            sql, params = self.table._format_query('select %s from %s where \
                                                   %s=%%s limit 1'
                                                   % (self.table.sourceSQL,
                                                      self.table.name,
                                                      self.table.sourceSQL),
                                                   (self.fromNode, ))
            self.table.cursor.execute(sql, params)
            if len(self.table.cursor.fetchall())<1:
                raise KeyError('node not in graph!')

    def __getitem__(self, target):
        sql, params = self.table._format_query('select %s from %s where \
                                               %s=%%s and %s=%%s limit 2'
                                               % (self.table.edgeSQL,
                                                  self.table.name,
                                                  self.table.sourceSQL,
                                                  self.table.targetSQL),
                                               (self.fromNode,
                                               self.table.pack_target(target)))
        self.table.cursor.execute(sql, params)
        l = self.table.cursor.fetchmany(2) # get at most two rows
        if len(l) != 1:
            raise KeyError('either no edge from source to target \
                           or not unique!')
        try:
            return self.table.unpack_edge(l[0][0]) # RETURN EDGE
        except IndexError:
            raise KeyError('no edge from node to target')

    def __setitem__(self, target, edge):
        sql, params = self.table._format_query('replace into %s values \
                                               (%%s,%%s,%%s)'
                                               % self.table.name,
                                               (self.fromNode,
                                                self.table.pack_target(target),
                                                self.table.pack_edge(edge)))
        self.table.cursor.execute(sql, params)
        if not hasattr(self.table, 'sourceDB') or \
           (hasattr(self.table, 'targetDB') and
            self.table.sourceDB is self.table.targetDB):
            self.table += target # ADD AS NODE TO GRAPH

    def __iadd__(self, target):
        self[target] = None
        return self # iadd MUST RETURN self!

    def __delitem__(self, target):
        sql, params = self.table._format_query('delete from %s where %s=%%s \
                                               and %s=%%s'
                                               % (self.table.name,
                                                  self.table.sourceSQL,
                                                  self.table.targetSQL),
                                               (self.fromNode,
                                               self.table.pack_target(target)))
        self.table.cursor.execute(sql, params)
        if self.table.cursor.rowcount < 1: # no rows deleted?
            raise KeyError('no edge from node to target')

    def iterator_query(self):
        sql, params = self.table._format_query('select %s,%s from %s where \
                                               %s=%%s and %s is not null'
                                               % (self.table.targetSQL,
                                                  self.table.edgeSQL,
                                                  self.table.name,
                                                  self.table.sourceSQL,
                                                  self.table.targetSQL),
                                               (self.fromNode, ))
        self.table.cursor.execute(sql, params)
        return self.table.cursor.fetchall()

    def keys(self):
        return [self.table.unpack_target(target_id)
                for target_id, edge_id in self.iterator_query()]

    def values(self):
        return [self.table.unpack_edge(edge_id)
                for target_id, edge_id in self.iterator_query()]

    def edges(self):
        return [(self.table.unpack_source(self.fromNode),
                 self.table.unpack_target(target_id),
                 self.table.unpack_edge(edge_id))
                for target_id, edge_id in self.iterator_query()]

    def items(self):
        return [(self.table.unpack_target(target_id),
                 self.table.unpack_edge(edge_id))
                for target_id, edge_id in self.iterator_query()]

    def __iter__(self):
        return iter(self.keys())

    def itervalues(self):
        return iter(self.values())

    def iteritems(self):
        return iter(self.items())

    def __len__(self):
        return len(self.keys())

    __cmp__ = graph_cmp


class SQLEdgelessDict(SQLEdgeDict):
    'for SQLGraph tables that lack edge_id column'

    def __getitem__(self, target):
        sql, params = self.table._format_query('select %s from %s where \
                                               %s=%%s and %s=%%s limit 2'
                                               % (self.table.targetSQL,
                                                  self.table.name,
                                                  self.table.sourceSQL,
                                                  self.table.targetSQL),
                                               (self.fromNode,
                                               self.table.pack_target(target)))
        self.table.cursor.execute(sql, params)
        l = self.table.cursor.fetchmany(2)
        if len(l) != 1:
            raise KeyError('either no edge from source to target \
                           or not unique!')
        return None # no edge info!

    def iterator_query(self):
        sql, params = self.table._format_query('select %s from %s where \
                                               %s=%%s and %s is not null'
                                               % (self.table.targetSQL,
                                                  self.table.name,
                                                  self.table.sourceSQL,
                                                  self.table.targetSQL),
                                               (self.fromNode, ))
        self.table.cursor.execute(sql, params)
        return [(t[0], None) for t in self.table.cursor.fetchall()]


SQLEdgeDict._edgelessClass = SQLEdgelessDict


class SQLGraphEdgeDescriptor(object):
    'provide an SQLEdges interface on demand'

    def __get__(self, obj, objtype):
        try:
            attrAlias = obj.attrAlias.copy()
        except AttributeError:
            return SQLEdges(obj.name, obj.cursor, graph=obj)
        else:
            return SQLEdges(obj.name, obj.cursor, attrAlias=attrAlias,
                            graph=obj)


def getColumnTypes(createTable, attrAlias={}, defaultColumnType='int',
                  columnAttrs=('source', 'target', 'edge'), **kwargs):
    'return list of [(colname, coltype), ...] for source, target, edge'
    l = []
    for attr in columnAttrs:
        try:
            attrName = attrAlias[attr + '_id']
        except KeyError:
            attrName = attr + '_id'
        try: # SEE IF USER SPECIFIED A DESIRED TYPE
            l.append((attrName, createTable[attr + '_id']))
            continue
        except (KeyError, TypeError):
            pass
        try: # get type info from primary key for that database
            db = kwargs[attr + 'DB']
            if db is None:
                raise KeyError # FORCE IT TO USE DEFAULT TYPE
        except KeyError:
            pass
        else: # INFER THE COLUMN TYPE FROM THE ASSOCIATED DATABASE KEYS...
            it = iter(db)
            try: # GET ONE IDENTIFIER FROM THE DATABASE
                k = it.next()
            except StopIteration:
                # Table is empty, read the SQL type from db.
                try:
                    l.append((attrName, db.columnType[db.primary_key]))
                    continue
                except AttributeError:
                    pass
            else: # GET THE TYPE FROM THIS IDENTIFIER
                if isinstance(k, int) or isinstance(k, long):
                    l.append((attrName, 'int'))
                    continue
                elif isinstance(k, str):
                    l.append((attrName, 'varchar(32)'))
                    continue
                else:
                    raise ValueError('SQLGraph node/edge must be int or str!')
        l.append((attrName, defaultColumnType))
        logger.warn('no type info found for %s, so using default: %s'
                    % (attrName, defaultColumnType))
    return l


class SQLGraph(SQLTableMultiNoCache):
    '''provide a graph interface via a SQL table.  Key capabilities are:
       - setitem with an empty dictionary: a dummy operation
       - getitem with a key that exists: return a placeholder
       - setitem with non empty placeholder: again a dummy operation
       EXAMPLE TABLE SCHEMA:
       create table mygraph (source_id int not null,target_id int,edge_id int,
              unique(source_id,target_id));
       '''
    _distinct_key = 'source_id'
    _pickleAttrs = SQLTableMultiNoCache._pickleAttrs.copy()
    _pickleAttrs.update(dict(sourceDB=0, targetDB=0, edgeDB=0,
                             allowMissingNodes=0))
    _edgeClass = SQLEdgeDict

    def __init__(self, name, *l, **kwargs):
        graphArgs, tableArgs = split_kwargs(kwargs,
                    ('attrAlias', 'defaultColumnType', 'columnAttrs',
                     'sourceDB', 'targetDB', 'edgeDB', 'simpleKeys',
                     'unpack_edge', 'edgeDictClass', 'graph'))
        if 'createTable' in kwargs: # CREATE A SCHEMA FOR THIS TABLE
            c = getColumnTypes(**kwargs)
            tableArgs['createTable'] = \
              'create table %s (%s %s not null,%s %s,%s %s,unique(%s,%s))' \
              % (name, c[0][0], c[0][1], c[1][0], c[1][1], c[2][0], c[2][1],
                 c[0][0], c[1][0])
        try:
            self.allowMissingNodes = kwargs['allowMissingNodes']
        except KeyError:
            pass
        SQLTableMultiNoCache.__init__(self, name, *l, **tableArgs)
        self.sourceSQL = self._attrSQL('source_id')
        self.targetSQL = self._attrSQL('target_id')
        try:
            self.edgeSQL = self._attrSQL('edge_id')
        except AttributeError:
            self.edgeSQL = None
            self._edgeClass = self._edgeClass._edgelessClass
        save_graph_db_refs(self, **kwargs)

    def __getitem__(self, k):
        return self._edgeClass(self.pack_source(k), self)

    def __iadd__(self, k):
        sql, params = self._format_query('delete from %s where %s=%%s and %s \
                                         is null' % (self.name, self.sourceSQL,
                                                     self.targetSQL),
                                         (self.pack_source(k), ))
        self.cursor.execute(sql, params)
        sql, params = self._format_query('insert %%(IGNORE)s into %s values \
                                         (%%s,NULL,NULL)' % self.name,
                                         (self.pack_source(k), ))
        self.cursor.execute(sql, params)
        return self # iadd MUST RETURN SELF!

    def __isub__(self, k):
        sql, params = self._format_query('delete from %s where %s=%%s'
                                         % (self.name, self.sourceSQL),
                                         (self.pack_source(k), ))
        self.cursor.execute(sql, params)
        if self.cursor.rowcount == 0:
            raise KeyError('node not found in graph')
        return self # iadd MUST RETURN SELF!

    __setitem__ = graph_setitem

    def __contains__(self, k):
        sql, params = self._format_query('select * from %s where %s=%%s \
                                         limit 1' % (self.name,
                                                     self.sourceSQL),
                                         (self.pack_source(k), ))
        self.cursor.execute(sql, params)
        l = self.cursor.fetchmany(2)
        return len(l) > 0

    def __invert__(self):
        'get an interface to the inverse graph mapping'
        try: # CACHED
            return self._inverse
        except AttributeError: # CONSTRUCT INTERFACE TO INVERSE MAPPING
            attrAlias = dict(source_id=self.targetSQL, # SWAP SOURCE & TARGET
                             target_id=self.sourceSQL,
                             edge_id=self.edgeSQL)
            if self.edgeSQL is None: # no edge interface
                del attrAlias['edge_id']
            self._inverse = SQLGraph(self.name, self.cursor,
                                     attrAlias=attrAlias,
                                     **graph_db_inverse_refs(self))
            self._inverse._inverse = self
            return self._inverse

    def __iter__(self):
        for k in SQLTableMultiNoCache.__iter__(self):
            yield self.unpack_source(k)

    def iteritems(self):
        for k in SQLTableMultiNoCache.__iter__(self):
            yield (self.unpack_source(k), self._edgeClass(k, self))

    def itervalues(self):
        for k in SQLTableMultiNoCache.__iter__(self):
            yield self._edgeClass(k, self)

    def keys(self):
        return [self.unpack_source(k) for k in SQLTableMultiNoCache.keys(self)]

    def values(self):
        return list(self.itervalues())

    def items(self):
        return list(self.iteritems())

    edges=SQLGraphEdgeDescriptor()
    update = update_graph

    def __len__(self):
        'get number of source nodes in graph'
        self.cursor.execute('select count(distinct %s) from %s'
                            % (self.sourceSQL, self.name))
        return self.cursor.fetchone()[0]

    __cmp__ = graph_cmp
    override_rich_cmp(locals()) # MUST OVERRIDE __eq__ ETC. TO USE OUR __cmp__!

##     def __cmp__(self, other):
##         node = ()
##         n = 0
##         d = None
##         it = iter(self.edges)
##         while True:
##             try:
##                 source, target, edge = it.next()
##             except StopIteration:
##                 source = None
##             if source != node:
##                 if d is not None:
##                     diff = cmp(n_target, len(d))
##                     if diff != 0:
##                         return diff
##                 if source is None:
##                     break
##                 node = source
##                 n += 1 # COUNT SOURCE NODES
##                 n_target = 0
##                 try:
##                     d = other[node]
##                 except KeyError:
##                     return 1
##             try:
##                 diff = cmp(edge, d[target])
##             except KeyError:
##                 return 1
##             if diff != 0:
##                 return diff
##             n_target += 1 # COUNT TARGET NODES FOR THIS SOURCE
##         return cmp(n, len(other))

    add_standard_packing_methods(locals())  ############ PACK / UNPACK METHODS


class SQLIDGraph(SQLGraph):
    add_trivial_packing_methods(locals())

SQLGraph._IDGraphClass = SQLIDGraph


class SQLEdgeDictClustered(dict):
    'simple cache for 2nd level dictionary of target_id:edge_id'

    def __init__(self, g, fromNode):
        self.g = g
        self.fromNode = fromNode
        dict.__init__(self)

    def __iadd__(self, l):
        for target_id, edge_id in l:
            dict.__setitem__(self, target_id, edge_id)
        return self # iadd MUST RETURN SELF!


class SQLEdgesClusteredDescr(object):

    def __get__(self, obj, objtype):
        e = SQLEdgesClustered(obj.table, obj.edge_id, obj.source_id,
                              obj.target_id, graph=obj,
                              **graph_db_inverse_refs(obj, True))
        for source_id, d in obj.d.iteritems(): # COPY EDGE CACHE
            e.load([(edge_id, source_id, target_id)
                    for (target_id, edge_id) in d.iteritems()])
        return e


class SQLGraphClustered(object):
    'SQL graph with clustered caching -- loads an entire cluster at a time'
    _edgeDictClass = SQLEdgeDictClustered

    def __init__(self, table, source_id='source_id', target_id='target_id',
                 edge_id='edge_id', clusterKey=None, **kwargs):
        import types
        if isinstance(table, types.StringType): # CREATE THE TABLE INTERFACE
            if clusterKey is None:
                raise ValueError('you must provide a clusterKey argument!')
            if 'createTable' in kwargs: # CREATE A SCHEMA FOR THIS TABLE
                c = getColumnTypes(attrAlias=dict(source_id=source_id,
                                                  target_id=target_id,
                                                  edge_id=edge_id), **kwargs)
                kwargs['createTable'] = 'create table %s (%s %s not null,%s \
                        %s,%s %s,unique(%s,%s))' % (table, c[0][0], c[0][1],
                                                    c[1][0], c[1][1], c[2][0],
                                                    c[2][1], c[0][0], c[1][0])
            table = SQLTableClustered(table, clusterKey=clusterKey, **kwargs)
        self.table = table
        self.source_id = source_id
        self.target_id = target_id
        self.edge_id = edge_id
        self.d = {}
        save_graph_db_refs(self, **kwargs)

    _pickleAttrs = dict(table=0, source_id=0, target_id=0, edge_id=0,
                        sourceDB=0, targetDB=0, edgeDB=0)

    def __getstate__(self):
        state = standard_getstate(self)
        state['d'] = {} # UNPICKLE SHOULD RESTORE GRAPH WITH EMPTY CACHE
        return state

    def __getitem__(self, k):
        'get edgeDict for source node k, from cache or by loading its cluster'
        try: # GET DIRECTLY FROM CACHE
            return self.d[k]
        except KeyError:
            if hasattr(self, '_isLoaded'):
                raise # ENTIRE GRAPH LOADED, SO k REALLY NOT IN THIS GRAPH
        # HAVE TO LOAD THE ENTIRE CLUSTER CONTAINING THIS NODE
        sql, params = self.table._format_query('select t2.%s,t2.%s,t2.%s \
               from %s t1,%s t2 where t1.%s=%%s and t1.%s=t2.%s group by t2.%s'
                                  % (self.source_id, self.target_id,
                                     self.edge_id, self.table.name,
                                     self.table.name, self.source_id,
                                     self.table.clusterKey,
                                     self.table.clusterKey,
                                     self.table.primary_key),
                                  (self.pack_source(k), ))
        self.table.cursor.execute(sql, params)
        self.load(self.table.cursor.fetchall()) # CACHE THIS CLUSTER
        return self.d[k] # RETURN EDGE DICT FOR THIS NODE

    def load(self, l=None, unpack=True):
        'load the specified rows (or all, if None provided) into local cache'
        if l is None:
            try: # IF ALREADY LOADED, NO NEED TO DO ANYTHING
                return self._isLoaded
            except AttributeError:
                pass
            self.table.cursor.execute('select %s,%s,%s from %s'
                                      % (self.source_id, self.target_id,
                                         self.edge_id, self.table.name))
            l = self.table.cursor.fetchall()
            self._isLoaded = True
            # Clear our cache as load() will replicate everything.
            self.d.clear()
        for source, target, edge in l: # SAVE TO OUR CACHE
            if unpack:
                source = self.unpack_source(source)
                target = self.unpack_target(target)
                edge = self.unpack_edge(edge)
            try:
                self.d[source] += [(target, edge)]
            except KeyError:
                d = self._edgeDictClass(self, source)
                d += [(target, edge)]
                self.d[source] = d

    def __invert__(self):
        'interface to reverse graph mapping'
        try:
            return self._inverse # INVERSE MAP ALREADY EXISTS
        except AttributeError:
            pass
        # JUST CREATE INTERFACE WITH SWAPPED TARGET & SOURCE
        self._inverse = SQLGraphClustered(self.table, self.target_id,
                                          self.source_id, self.edge_id,
                                          **graph_db_inverse_refs(self))
        self._inverse._inverse = self
        for source, d in self.d.iteritems(): # INVERT OUR CACHE
            self._inverse.load([(target, source, edge)
                                for (target, edge) in d.iteritems()],
                               unpack=False)
        return self._inverse
    edges=SQLEdgesClusteredDescr() # CONSTRUCT EDGE INTERFACE ON DEMAND
    update = update_graph
    add_standard_packing_methods(locals())  ############ PACK / UNPACK METHODS

    def __iter__(self): ################# ITERATORS
        'uses db select; does not force load'
        return iter(self.keys())

    def keys(self):
        'uses db select; does not force load'
        self.table.cursor.execute('select distinct(%s) from %s'
                                  % (self.source_id, self.table.name))
        return [self.unpack_source(t[0])
                for t in self.table.cursor.fetchall()]

    methodFactory(['iteritems', 'items', 'itervalues', 'values'],
                  'lambda self: (self.load(), self.d.%s())[1]', locals())

    def __contains__(self, k):
        try:
            x = self[k]
            return True
        except KeyError:
            return False


class SQLIDGraphClustered(SQLGraphClustered):
    add_trivial_packing_methods(locals())

SQLGraphClustered._IDGraphClass = SQLIDGraphClustered


class SQLEdgesClustered(SQLGraphClustered):
    'edges interface for SQLGraphClustered'
    _edgeDictClass = list
    _pickleAttrs = SQLGraphClustered._pickleAttrs.copy()
    _pickleAttrs.update(dict(graph=0))

    def keys(self):
        self.load()
        result = []
        for edge_id, l in self.d.iteritems():
            for source_id, target_id in l:
                result.append((self.graph.unpack_source(source_id),
                               self.graph.unpack_target(target_id),
                               self.graph.unpack_edge(edge_id)))
        return result


class ForeignKeyInverse(object):
    'map each key to a single value according to its foreign key'

    def __init__(self, g):
        self.g = g

    def __getitem__(self, obj):
        self.check_obj(obj)
        source_id = getattr(obj, self.g.keyColumn)
        if source_id is None:
            return None
        return self.g.sourceDB[source_id]

    def __setitem__(self, obj, source):
        self.check_obj(obj)
        if source is not None:
            # Ensures performing all the right caching operations.
            self.g[source][obj] = None
        else: # DELETE PRE-EXISTING EDGE IF PRESENT
            if not hasattr(obj, '_localOnly'):
                # Only cache, don't save to database.
                old_source = self[obj]
                if old_source is not None:
                    del self.g[old_source][obj]

    def check_obj(self, obj):
        'raise KeyError if obj not from this db'
        try:
            if obj.db is not self.g.targetDB:
                raise AttributeError
        except AttributeError:
            raise KeyError('key is not from targetDB of this graph!')

    def __contains__(self, obj):
        try:
            self.check_obj(obj)
            return True
        except KeyError:
            return False

    def __iter__(self):
        return self.g.targetDB.itervalues()

    def keys(self):
        return self.g.targetDB.values()

    def iteritems(self):
        for obj in self:
            source_id = getattr(obj, self.g.keyColumn)
            if source_id is None:
                yield obj, None
            else:
                yield obj, self.g.sourceDB[source_id]

    def items(self):
        return list(self.iteritems())

    def itervalues(self):
        for obj, val in self.iteritems():
            yield val

    def values(self):
        return list(self.itervalues())

    def __invert__(self):
        return self.g


class ForeignKeyEdge(dict):
    '''edge interface to a foreign key in an SQL table.
Caches dict of target nodes in itself; provides dict interface.
Adds or deletes edges by setting foreign key values in the table'''

    def __init__(self, g, k):
        dict.__init__(self)
        self.g = g
        self.src = k
        for v in g.targetDB.select('where %s=%%s' % g.keyColumn, (k.id, )):
            dict.__setitem__(self, v, None) # SAVE IN CACHE

    def __setitem__(self, dest, v):
        if not hasattr(dest, 'db') or dest.db is not self.g.targetDB:
            raise KeyError('dest is not in the targetDB bound to this graph!')
        if v is not None:
            raise ValueError('sorry,this graph cannot store edge information!')
        if not hasattr(dest, '_localOnly'):
            # Only cache, don't save to database.
            old_source = self.g._inverse[dest] # CHECK FOR PRE-EXISTING EDGE
            if old_source is not None: # REMOVE OLD EDGE FROM CACHE
                dict.__delitem__(self.g[old_source], dest)
        #self.g.targetDB._update(dest.id, self.g.keyColumn, self.src.id) # SAVE TO DB
        setattr(dest, self.g.keyColumn, self.src.id) # SAVE TO DB ATTRIBUTE
        dict.__setitem__(self, dest, None) # SAVE IN CACHE

    def __delitem__(self, dest):
        #self.g.targetDB._update(dest.id, self.g.keyColumn, None) # REMOVE FOREIGN KEY VALUE
        setattr(dest, self.g.keyColumn, None) # SAVE TO DB ATTRIBUTE
        dict.__delitem__(self, dest) # REMOVE FROM CACHE


class ForeignKeyGraph(object, UserDict.DictMixin):
    '''graph interface to a foreign key in an SQL table
Caches dict of target nodes in itself; provides dict interface.
    '''

    def __init__(self, sourceDB, targetDB, keyColumn, autoGC=True, **kwargs):
        '''sourceDB is any database of source nodes;
        targetDB must be an SQL database of target nodes;
        keyColumn is the foreign key column name in targetDB
        for looking up sourceDB IDs.'''
        if autoGC: # automatically garbage collect unused objects
            self._weakValueDict = RecentValueDictionary(autoGC) # object cache
        else:
            self._weakValueDict = {}
        self.autoGC = autoGC
        self.sourceDB = sourceDB
        self.targetDB = targetDB
        self.keyColumn = keyColumn
        self._inverse = ForeignKeyInverse(self)

    _pickleAttrs = dict(sourceDB=0, targetDB=0, keyColumn=0, autoGC=0)
    __getstate__ = standard_getstate ########### SUPPORT FOR PICKLING
    __setstate__ = standard_setstate

    def _inverse_schema(self):
        '''Provide custom schema rule for inverting this graph...
        Just use keyColumn!'''
        return dict(invert=True, uniqueMapping=True)

    def __getitem__(self, k):
        if not hasattr(k, 'db') or k.db is not self.sourceDB:
            raise KeyError('object is not in the sourceDB bound \
                           to this graph!')
        try:
            return self._weakValueDict[k.id] # get from cache
        except KeyError:
            pass
        d = ForeignKeyEdge(self, k)
        self._weakValueDict[k.id] = d # save in cache
        return d

    def __setitem__(self, k, v):
        raise KeyError('''do not save as g[k]=v.  Instead follow a graph
interface: g[src]+=dest, or g[src][dest]=None (no edge info allowed)''')

    def __delitem__(self, k):
        raise KeyError('''Instead of del g[k], follow a graph
interface: del g[src][dest]''')

    def keys(self):
        return self.sourceDB.values()

    __invert__ = standard_invert


def describeDBTables(name, cursor, idDict):
    """
    Get table info about database <name> via <cursor>, and store primary keys
    in idDict, along with a list of the tables each key indexes.
    """
    cursor.execute('use %s' % name)
    cursor.execute('show tables')
    tables = {}
    l = [c[0] for c in cursor.fetchall()]
    for t in l:
        tname = name + '.' + t
        o = SQLTable(tname, cursor)
        tables[tname] = o
        for f in o.description:
            if f == o.primary_key:
                idDict.setdefault(f, []).append(o)
            elif f[-3:] == '_id' and f not in idDict:
                idDict[f] = []
    return tables


def indexIDs(tables, idDict=None):
    "Get an index of primary keys in the <tables> dictionary."
    if idDict == None:
        idDict = {}
    for o in tables.values():
        if o.primary_key:
            # Maintain a list of tables with this primary key.
            if o.primary_key not in idDict:
                idDict[o.primary_key] = []
            idDict[o.primary_key].append(o)
        for f in o.description:
            if f[-3:] == '_id' and f not in idDict:
                idDict[f] = []
    return idDict


def suffixSubset(tables, suffix):
    "Filter table index for those matching a specific suffix"
    subset = {}
    for name, t in tables.items():
        if name.endswith(suffix):
            subset[name] = t
    return subset


PRIMARY_KEY=1


def graphDBTables(tables, idDict):
    g = dictgraph()
    for t in tables.values():
        for f in t.description:
            if f == t.primary_key:
                edgeInfo = PRIMARY_KEY
            else:
                edgeInfo = None
            g.setEdge(f, t, edgeInfo)
            g.setEdge(t, f, edgeInfo)
    return g


SQLTypeTranslation = {types.StringType: 'varchar(32)',
                      types.IntType: 'int',
                      types.FloatType: 'float'}


def createTableFromRepr(rows, tableName, cursor, typeTranslation=None,
                        optionalDict=None, indexDict=()):
    """Save rows into SQL tableName using cursor, with optional
       translations of columns to specific SQL types (specified
       by typeTranslation dict).
       - optionDict can specify columns that are allowed to be NULL.
       - indexDict can specify columns that must be indexed; columns
       whose names end in _id will be indexed by default.
       - rows must be an iterator which in turn returns dictionaries,
       each representing a tuple of values (indexed by their column
       names).
    """
    try:
        row = rows.next() # GET 1ST ROW TO EXTRACT COLUMN INFO
    except StopIteration:
        return # IF rows EMPTY, NO NEED TO SAVE ANYTHING, SO JUST RETURN
    try:
        createTableFromRow(cursor, tableName, row, typeTranslation,
                           optionalDict, indexDict)
    except:
        pass
    storeRow(cursor, tableName, row) # SAVE OUR FIRST ROW
    for row in rows: # NOW SAVE ALL THE ROWS
        storeRow(cursor, tableName, row)


def createTableFromRow(cursor, tableName, row, typeTranslation=None,
                        optionalDict=None, indexDict=()):
    create_defs = []
    for col, val in row.items(): # PREPARE SQL TYPES FOR COLUMNS
        coltype = None
        if typeTranslation != None and col in typeTranslation:
            coltype = typeTranslation[col] # USER-SUPPLIED TRANSLATION
        elif type(val) in SQLTypeTranslation:
            coltype = SQLTypeTranslation[type(val)]
        else: # SEARCH FOR A COMPATIBLE TYPE
            for t in SQLTypeTranslation:
                if isinstance(val, t):
                    coltype = SQLTypeTranslation[t]
                    break
        if coltype == None:
            raise TypeError("Don't know SQL type to use for %s" % col)
        create_def = '%s %s' % (col, coltype)
        if optionalDict == None or col not in optionalDict:
            create_def += ' not null'
        create_defs.append(create_def)
    for col in row: # CREATE INDEXES FOR ID COLUMNS
        if col[-3:] == '_id' or col in indexDict:
            create_defs.append('index(%s)' % col)
    cmd = 'create table if not exists %s (%s)' % (tableName,
                                                  ','.join(create_defs))
    cursor.execute(cmd) # CREATE THE TABLE IN THE DATABASE


def storeRow(cursor, tableName, row):
    row_format = ','.join(len(row) * ['%s'])
    cmd = 'insert into %s values (%s)' % (tableName, row_format)
    cursor.execute(cmd, tuple(row.values()))


def storeRowDelayed(cursor, tableName, row):
    row_format = ','.join(len(row) * ['%s'])
    cmd = 'insert delayed into %s values (%s)' % (tableName, row_format)
    cursor.execute(cmd, tuple(row.values()))


class TableGroup(dict):
    'provide attribute access to dbname qualified tablenames'

    def __init__(self, db='test', suffix=None, **kw):
        dict.__init__(self)
        self.db=db
        if suffix is not None:
            self.suffix=suffix
        for k, v in kw.items():
            if v is not None and '.' not in v:
                v=self.db+'.'+v  # ADD DATABASE NAME AS PREFIX
            self[k]=v

    def __getattr__(self, k):
        return self[k]


def sqlite_connect(*args, **kwargs):
    sqlite = import_sqlite()
    connection = sqlite.connect(*args, **kwargs)
    cursor = connection.cursor()
    return connection, cursor


class DBServerInfo(object):
    'picklable reference to a database server'

    def __init__(self, moduleName='MySQLdb', serverSideCursors=False,
                 blockIterators=True, *args, **kwargs):
        try:
            self.__class__ = _DBServerModuleDict[moduleName]
        except KeyError:
            raise ValueError('Module name not found in _DBServerModuleDict: '\
                             + moduleName)
        self.moduleName = moduleName
        self.args = args  # connection arguments
        self.kwargs = kwargs
        self.serverSideCursors = serverSideCursors
        self.custom_iter_keys = blockIterators
        if self.serverSideCursors and not self.custom_iter_keys:
            raise ValueError('serverSideCursors=True requires \
                             blockIterators=True!')

    def cursor(self):
        """returns cursor associated with the DB server info (reused)"""
        try:
            return self._cursor
        except AttributeError:
            self._start_connection()
            return self._cursor

    def new_cursor(self, arraysize=None):
        """returns a NEW cursor; you must close it yourself! """
        if not hasattr(self, '_connection'):
            self._start_connection()
        cursor = self._connection.cursor()
        if arraysize is not None:
            cursor.arraysize = arraysize
        return cursor

    def close(self):
        """Close file containing this database"""
        self._cursor.close()
        self._connection.close()
        del self._cursor
        del self._connection

    def __getstate__(self):
        """return all picklable arguments"""
        return dict(args=self.args, kwargs=self.kwargs,
                    moduleName=self.moduleName,
                    serverSideCursors=self.serverSideCursors,
                    custom_iter_keys=self.custom_iter_keys)


class MySQLServerInfo(DBServerInfo):
    'customized for MySQLdb SSCursor support via new_cursor()'
    _serverType = 'mysql'

    def _start_connection(self):
        self._connection, self._cursor = mysql_connect(*self.args,
                                                       **self.kwargs)

    def new_cursor(self, arraysize=None):
        'provide streaming cursor support'
        if not self.serverSideCursors: # use regular MySQLdb cursor
            return DBServerInfo.new_cursor(self, arraysize)
        try:
            conn = self._conn_sscursor
        except AttributeError:
            self._conn_sscursor, cursor = mysql_connect(useStreaming=True,
                                                        *self.args,
                                                        **self.kwargs)
        else:
            cursor = self._conn_sscursor.cursor()
        if arraysize is not None:
            cursor.arraysize = arraysize
        return cursor

    def close(self):
        DBServerInfo.close(self)
        try:
            self._conn_sscursor.close()
            del self._conn_sscursor
        except AttributeError:
            pass

    def iter_keys(self, db, cursor, map_f=iter,
                  cache_f=lambda x: [t[0] for t in x], **kwargs):
        block_iterator = BlockIterator(db, cursor, **kwargs)
        try:
            cache_f = block_iterator.cache_f
        except AttributeError:
            pass
        return db.generic_iterator(cursor=cursor, cache_f=cache_f,
                                   map_f=map_f, fetch_f=block_iterator)


class CursorCloser(object):
    """container for ensuring cursor.close() is called, when this obj deleted.
    For Python 2.5+, we could replace this with a try... finally clause
    in a generator function such as generic_iterator(); see PEP 342 or
    What's New in Python 2.5.  """

    def __init__(self, cursor):
        self.cursor = cursor

    def __del__(self):
        self.cursor.close()


class BlockIterator(CursorCloser):
    'workaround for MySQLdb iteration horrible performance'

    def __init__(self, db, cursor, selectCols, whereClause='', **kwargs):
        self.db = db
        self.cursor = cursor
        self.selectCols = selectCols
        self.kwargs = kwargs
        self.whereClause = ''
        if kwargs['orderBy']: # use iterSQL/iterColumns for WHERE / SELECT
            self.whereSQL = db.iterSQL
            if selectCols == '*': # extracting all columns
                self.whereParams = [db.data[col] for col in db.iterColumns]
            else: # selectCols is single column
                iterColumns = list(db.iterColumns)
                try: # if selectCols in db.iterColumns, just use that
                    i = iterColumns.index(selectCols)
                except ValueError: # have to append selectCols
                    i = len(db.iterColumns)
                    iterColumns += [selectCols]
                self.selectCols = ','.join(iterColumns)
                self.whereParams = range(len(db.iterColumns))
                if i > 0: # need to extract desired column
                    self.cache_f = lambda x: [t[i] for t in x]
        else: # just use primary key
            self.whereSQL = 'WHERE %s>%%s' % db.primary_key
            self.whereParams = (db.data[db.primary_key],)
        self.params = ()
        self.done = False

    def __call__(self):
        'get the next block of data'
        if self.done:
            return ()
        self.db._select(self.whereClause, self.params, cursor=self.cursor,
                        limit='LIMIT %s' % self.cursor.arraysize,
                        selectCols=self.selectCols, **(self.kwargs))
        rows = self.cursor.fetchall()
        if len(rows) < self.cursor.arraysize: # iteration complete
            self.done = True
            return rows
        lastrow = rows[-1] # extract params from the last row in this block
        if len(lastrow) > 1:
            self.params = [lastrow[icol] for icol in self.whereParams]
        else:
            self.params = lastrow
        self.whereClause = self.whereSQL
        return rows


class SQLiteServerInfo(DBServerInfo):
    """picklable reference to a sqlite database"""
    _serverType = 'sqlite'

    def __init__(self, database, *args, **kwargs):
        """Takes same arguments as sqlite3.connect()"""
        DBServerInfo.__init__(self, 'sqlite3',  # save abs path!
                              database=SourceFileName(database),
                              *args, **kwargs)

    def _start_connection(self):
        self._connection, self._cursor = sqlite_connect(*self.args,
                                                        **self.kwargs)

    def __getstate__(self):
        database = self.kwargs.get('database', False) or self.args[0]
        if database == ':memory:':
            raise ValueError('SQLite in-memory database is not picklable!')
        return DBServerInfo.__getstate__(self)

# list of DBServerInfo subclasses for different modules
_DBServerModuleDict = dict(MySQLdb=MySQLServerInfo,
                           sqlite3=SQLiteServerInfo)


class MapView(object, UserDict.DictMixin):
    'general purpose 1:1 mapping defined by any SQL query'

    def __init__(self, sourceDB, targetDB, viewSQL, cursor=None,
                 serverInfo=None, inverseSQL=None, **kwargs):
        self.sourceDB = sourceDB
        self.targetDB = targetDB
        self.viewSQL = viewSQL
        self.inverseSQL = inverseSQL
        if cursor is None:
            if serverInfo is not None: # get cursor from serverInfo
                cursor = serverInfo.cursor()
            else:
                try: # can we get it from our other db?
                    serverInfo = sourceDB.serverInfo
                except AttributeError:
                    raise ValueError('you must provide serverInfo or cursor!')
                else:
                    cursor = serverInfo.cursor()
        self.cursor = cursor
        self.serverInfo = serverInfo
        self.get_sql_format(False) # get sql formatter for this db interface

    _schemaModuleDict = _schemaModuleDict # default module list
    get_sql_format = get_table_schema

    def __getitem__(self, k):
        if not hasattr(k, 'db') or k.db is not self.sourceDB:
            raise KeyError('object is not in the sourceDB bound to this map!')
        sql, params = self._format_query(self.viewSQL, (k.id, ))
        self.cursor.execute(sql, params) # formatted for this db interface
        t = self.cursor.fetchmany(2) # get at most two rows
        if len(t) != 1:
            raise KeyError('%s not found in MapView, or not unique'
                           % str(k))
        return self.targetDB[t[0][0]] # get the corresponding object

    _pickleAttrs = dict(sourceDB=0, targetDB=0, viewSQL=0, serverInfo=0,
                        inverseSQL=0)
    __getstate__ = standard_getstate
    __setstate__ = standard_setstate
    __setitem__ = __delitem__ = clear = pop = popitem = update = \
                  setdefault = read_only_error

    def __iter__(self):
        'only yield sourceDB items that are actually in this mapping!'
        for k in self.sourceDB.itervalues():
            try:
                self[k]
                yield k
            except KeyError:
                pass

    def keys(self):
        return [k for k in self] # don't use list(self); causes infinite loop!

    def __invert__(self):
        try:
            return self._inverse
        except AttributeError:
            if self.inverseSQL is None:
                raise ValueError('this MapView has no inverseSQL!')
            self._inverse = self.__class__(self.targetDB, self.sourceDB,
                                           self.inverseSQL, self.cursor,
                                           serverInfo=self.serverInfo,
                                           inverseSQL=self.viewSQL)
            self._inverse._inverse = self
            return self._inverse


class GraphViewEdgeDict(UserDict.DictMixin):
    'edge dictionary for GraphView: just pre-loaded on init'

    def __init__(self, g, k):
        self.g = g
        self.k = k
        sql, params = self.g._format_query(self.g.viewSQL, (k.id, ))
        self.g.cursor.execute(sql, params) # run the query
        l = self.g.cursor.fetchall() # get results
        if len(l) <= 0:
            raise KeyError('key %s not in GraphView' % k.id)
        self.targets = [t[0] for t in l] # preserve order of the results
        d = {} # also keep targetID:edgeID mapping
        if self.g.edgeDB is not None: # save with edge info
            for t in l:
                d[t[0]] = t[1]
        else:
            for t in l:
                d[t[0]] = None
        self.targetDict = d

    def __len__(self):
        return len(self.targets)

    def __iter__(self):
        for k in self.targets:
            yield self.g.targetDB[k]

    def keys(self):
        return list(self)

    def iteritems(self):
        if self.g.edgeDB is not None: # save with edge info
            for k in self.targets:
                yield (self.g.targetDB[k], self.g.edgeDB[self.targetDict[k]])
        else: # just save the list of targets, no edge info
            for k in self.targets:
                yield (self.g.targetDB[k], None)

    def __getitem__(self, o, exitIfFound=False):
        'for the specified target object, return its associated edge object'
        try:
            if o.db is not self.g.targetDB:
                raise KeyError('key is not part of targetDB!')
            edgeID = self.targetDict[o.id]
        except AttributeError:
            raise KeyError('key has no id or db attribute?!')
        if exitIfFound:
            return
        if self.g.edgeDB is not None: # return the edge object
            return self.g.edgeDB[edgeID]
        else: # no edge info
            return None

    def __contains__(self, o):
        try:
            self.__getitem__(o, True) # raise KeyError if not found
            return True
        except KeyError:
            return False

    __setitem__ = __delitem__ = clear = pop = popitem = update = \
                  setdefault = read_only_error


class GraphView(MapView):
    'general purpose graph interface defined by any SQL query'

    def __init__(self, sourceDB, targetDB, viewSQL, cursor=None, edgeDB=None,
                 **kwargs):
        '''if edgeDB not None, viewSQL query must return
        (targetID, edgeID) tuples'''
        self.edgeDB = edgeDB
        MapView.__init__(self, sourceDB, targetDB, viewSQL, cursor, **kwargs)

    def __getitem__(self, k):
        if not hasattr(k, 'db') or k.db is not self.sourceDB:
            raise KeyError('object is not in the sourceDB bound to this map!')
        return GraphViewEdgeDict(self, k)
    _pickleAttrs = MapView._pickleAttrs.copy()
    _pickleAttrs.update(dict(edgeDB=0))


class SQLSequence(SQLRow, SequenceBase):
    """Transparent access to a DB row representing a sequence.

    Use attrAlias dict to rename 'length' to something else.
    """

    def _init_subclass(cls, db, **kwargs):
        db.seqInfoDict = db # db will act as its own seqInfoDict
        SQLRow._init_subclass(db=db, **kwargs)
    _init_subclass = classmethod(_init_subclass)

    def __init__(self, id):
        SQLRow.__init__(self, id)
        SequenceBase.__init__(self)

    def __len__(self):
        return self.length

    def strslice(self, start, end):
        "Efficient access to slice of a sequence, useful for huge contigs"
        return self._select('%%(SUBSTRING)s(%s %%(SUBSTR_FROM)s %d \
                            %%(SUBSTR_FOR)s %d)' % (self.db._attrSQL('seq'),
                                                    start + 1, end - start))


class DNASQLSequence(SQLSequence):
    _seqtype=DNA_SEQTYPE


class RNASQLSequence(SQLSequence):
    _seqtype=RNA_SEQTYPE


class ProteinSQLSequence(SQLSequence):
    _seqtype=PROTEIN_SEQTYPE




import types


# STORES DICTIONARY OF ATTRIBUTE-BOUND GRAPHS
# AND LIST OF UNBOUND GRAPHS
class SchemaDict(dict):
    """Container for schema rules bound to a class or object. Rules
    are stored in two indexes for fast access, indexed by graph, and
    indexed by attrname. Use += and -= to add or remove rules.
    """

    def __init__(self, newlist=(), baselist=()):
        "Initialize schema list from list of base classes and newlist of rules"
        self.attrs = {}
        dict.__init__(self)
        # COMBINE SCHEMAS FROM PARENTS WITH NEW SCHEMA LIST
        for b in baselist:
            if hasattr(b, '__class_schema__'):
                self.update(b.__class_schema__)
                self.attrs.update(b.__class_schema__.attrs)
        for i in newlist: # newlist OVERRIDES OLD DEFS FROM baselist
            self += i

    def __iadd__(self, i):
        "Add a schema rule to this SchemaDict"
        g = i[0]
        if len(i) >= 2:
            if isinstance(i[1], types.StringType):
                if i[1] in self.attrs: # REMOVE OLD ENTRY
                    self -= self.attrs[i[1]]
                self.attrs[i[1]] = i # SAVE IN INDEX ACCORDING TO ATTR NAME
            else:
                raise TypeError('Attribute name must be a string')
        if g not in self:
            self[g] = []
        self[g].append(i) # SAVE IN GRAPH INDEX
        return self # REQUIRED FROM iadd()!!

    def __isub__(self, i):
        "Remove a schema rule from this SchemaDict"
        g = i[0]
        if g not in self:
            raise KeyError('graph not found in SchemaDict!')
        self[g].remove(i) # REMOVE OLD ENTRY
        if len(self[g]) == 0: # REMOVE EMPTY LIST
            del self[g]
        if len(i) >= 2:
            if isinstance(i[1], types.StringType):
                if i[1] not in self.attrs:
                    raise KeyError('attribute not found in SchemaDict!')
                del self.attrs[i[1]] # REMOVE OLD ENTRY
            else:
                raise TypeError('Attribute name must be a string')
        return self # REQUIRED FROM iadd()!!

    def initInstance(self, obj):
        "Add obj as new node to all graphs referenced by this SchemaDict."
        for g, l in self.items(): # GET ALL OUR RULES
            for s in l:
                if obj not in g:
                    g.__iadd__(obj, (s, ))  # ADD OBJECT TO GRAPH USING RULE s

    def getschema(self, attr=None, graph=None):
        "Return list of schema rules that match attr / graph arguments."
        if attr:
            if attr in self.attrs:
                return [self.attrs[attr]]
        elif graph:
            if graph in self:
                return self[graph]
        else:
            raise ValueError('You must specify an attribute or graph.')
        return [] # DIDN'T FIND ANYTHING


class SchemaList(list):
    "Temporary container for returned schema list, with attached methods"

    def __init__(self, obj):
        self.obj = obj # OBJECT THAT WE'RE DESCRIBING SCHEMA OF
        list.__init__(self) # CALL SUPERCLASS CONSTRUCTOR

    def __iadd__(self, rule):
        "Add a new schema rule to object described by this SchemaList"
        if not hasattr(self.obj, '__schema__'):
            self.obj.__schema__ = SchemaDict()
        self.obj.__schema__ += rule
        return self # REQUIRED FROM iadd()!!

    # PROBABLY NEED AN __isub__() TOO??


######################
# getschema, getnodes, getedges
# these functions are analogous to getattr, except they get graph information

def getschema(o, attr=None, graph=None):
    "Get list of schema rules for object o that match attr / graph arguments."
    found = SchemaList(o)
    attrs = {}
    if hasattr(o, '__schema__'):
        for s in o.__schema__.getschema(attr, graph):
            found.append(s)
            if isinstance(s[1], types.StringType):
                attrs[s[1]] = None
    if attr and len(found) > 0: # DON'T PROCEED
        return found
    if hasattr(o, '__class_schema__'):
        for s in o.__class_schema__.getschema(attr, graph):
            if not isinstance(s[1], types.StringType) or s[1] not in attrs:
                found.append(s) # DON'T OVERWRITE OBJECT __schema__ BINDINGS
    return found


def setschema(o, attr, graph):
    """Bind object to graph, and if attr not None, also bind graph
    to this attribute."""
    if not hasattr(o, '__schema__'):
        o.__schema__ = SchemaDict()
    o.__schema__ += (graph, attr)


def getnodes(o, attr=None, graph=None):
    """Get destination nodes from graph bindings of o, limited to the
    specific attribute or graph if specified"""
    if attr:
        if hasattr(o, '__schema__') and attr in o.__schema__:
            return getattr(o, o.__schema__[attr][2]) # RETURN THE PRODUCT

        if hasattr(o, '__class_schema__') and attr in o.__class_schema__:
            return getattr(o, o.__class_schema__[attr][2]) # RETURN THE PRODUCT
        raise AttributeError('No attribute named %s in object %s' % (attr, o))
    elif graph: # JUST LOOK UP THE GRAPH TRIVIALLY
        return graph[o]
    else: # SHOULD WE GET ALL NODES FROM ALL SCHEMA ENTRIES?  HOW??
        raise ValueError('You must pass an attribute name or graph')


def getedges(o, attr=None, graph=None):
    """Get edges from graph bindings of o, limited to the specific attribute
    or graph if specified"""
    g = getnodes(o, attr, graph) # CAN JUST REUSE THE LOGIC OF getnodes
    if g and hasattr(g, 'edges'):
        return g.edges()
    else:
        return None



import os
import types
import classutil
import logger
from UserDict import DictMixin


class NLMSASeqList(list):

    def __init__(self, nlmsaSeqDict):
        list.__init__(self)
        self.nlmsaSeqDict = nlmsaSeqDict

    def __getitem__(self, nlmsaID):
        'return NLMSASequence for a given nlmsa_id'
        try:
            return list.__getitem__(self, nlmsaID)
        except IndexError:
            seqID, nsID = self.nlmsaSeqDict.IDdict[str(nlmsaID)]
            return list.__getitem__(self, nsID)

    def getSeq(self, nlmsaID):
        'return seq for a given nlmsa_id'
        seqID, nsID = self.nlmsaSeqDict.IDdict[str(nlmsaID)]
        return self.nlmsaSeqDict.nlmsa.seqDict[seqID]

    def getSeqID(self, nlmsaID):
        'return seqID for a given nlmsa_id'
        seqID, nsID = self.nlmsaSeqDict.IDdict[str(nlmsaID)]
        return seqID

    def is_lpo(self, id):
        if id >= len(self):
            return False
        ns = self[id]
        if ns.is_lpo:
            return True
        else:
            return False

    def nextID(self):
        return len(self)


class EmptySliceError(KeyError):
    pass


class EmptyAlignmentError(ValueError):
    pass


class EmptySlice:
    'Empty slice for use by NLMSASlice'

    def __init__(self, seq):
        self.seq = seq

    def edges(self, *args, **kwargs):
        return []

    def items(self, **kwargs):
        return []

    def iteritems(self, **kwargs):
        return iter([])

    def keys(self, **kwargs):
        return []

    def __iter__(self):
        return iter([])

    def __getitem__(self, k):
        raise KeyError

    def __len__(self):
        return 0

    def matchIntervals(self, seq=None):
        return []

    def findSeqEnds(self, seq):
        raise KeyError('seq not aligned in this interval')

    def generateSeqEnds(self):
        return []

    def groupByIntervals(self, **kwargs):
        return {}

    def groupBySequences(self, **kwargs):
        return []

    def split(self, **kwargs):
        return []

    def regions(self, **kwargs):
        return []

    def __cmp__(self, other):
        return cmp(self.seq, other.seq)

    def rawIvals(self):
        return []


class _NLMSASeqDict_ValueWrapper(object):
    """A wrapper class for NLMSASeqDict to use to store 3-tuples in its cache.

    NLMSASeqDict has a most-recent-values cache containing (id,
    seqlist, offset) tuples for each referenced pathForward.  However,
    tuples cannot be stored in a weakref dictionary.  This class provides
    a tuple-like wrapper object that *can* be stored in a weakref dict.
    
    """
    def __init__(self, nlmsaID, seqlist, offset):
        self.v = (nlmsaID, seqlist, offset)

    def __hash__(self):
        return hash(self.v)

    def __len__(self):
        return 3

    def __getitem__(self, n):
        return self.v[n]

_DEFAULT_SEQUENCE_CACHE_SIZE=100
class NLMSASeqDict(object, DictMixin):
    """Index sequences by pathForward, and use list to keep reverse mapping.

    Keeps a cache of n most recently accessed sequences, up to
    maxSequenceCacheSize (defaults to 100).
    
    """

    def __init__(self, nlmsa, filename, mode, idDictClass=None,
                 maxSequenceCacheSize=_DEFAULT_SEQUENCE_CACHE_SIZE):
        self._cache = classutil.RecentValueDictionary(maxSequenceCacheSize)
        self.seqlist = NLMSASeqList(self)
        self.nlmsa = nlmsa
        self.filename = filename
        if mode == 'memory': # just use python dictionary
            idDictClass = dict
        elif mode == 'w': # new database
            mode = 'n'
        if idDictClass is None: # use persistent id dictionary storage
            self.seqIDdict = classutil.open_shelve(filename + '.seqIDdict',
                                                   mode)
            self.IDdict = classutil.open_shelve(filename + '.idDict', mode)
        else: # user supplied class for id dictionary storage
            self.seqIDdict = idDictClass()
            self.IDdict = idDictClass()

    def saveSeq(self, seq, nsID= -1, offset=0, nlmsaID=None):
        'save mapping of seq to specified (nlmsaID,ns,offset)'
        if nsID < 0: # let the union figure it out
            self.nlmsa.currentUnion.__iadd__(seq)
            return # the union added it for us, no need to do anything
        if isinstance(seq, types.StringType):
            id = seq # treat this as fully qualified identifier
        else: # get the identfier from the seq / database
            id = self.getSeqID(seq)
        if nlmsaID is None: # allocate a new unique id
            nlmsaID = self.nlmsa.nextID()
        self.seqIDdict[id] = nlmsaID, nsID, offset
        self.IDdict[str(nlmsaID)] = id, nsID

    def getIDcoords(self, seq):
        'return nlmsaID,start,stop for a given seq ival.'
        nlmsaID = self.getID(seq)
        return nlmsaID, seq.start, seq.stop # standard coords

    def getID(self, seq):
        'return nlmsa_id for a given seq'
        return self[seq][0]

    def __getitem__(self, seq):
        'return nlmsaID,NLMSASequence,offset for a given seq'
        if not hasattr(seq, 'annotationType'): # don't cache annotations
            try: # look in our sequence cache
                return self._cache[seq.pathForward]
            except AttributeError:
                raise KeyError('key must be a sequence interval!')
            except KeyError:
                pass
        seqID = self.getSeqID(seq) # use seq id to look up...
        try:
            nlmsaID, nsID, offset = self.seqIDdict[seqID]
        except KeyError:
            raise KeyError('seq not found in this alignment')
        v = nlmsaID, self.seqlist[nsID], offset
        if not hasattr(seq, 'annotationType'): # don't cache annotations
            self._cache[seq.pathForward] = _NLMSASeqDict_ValueWrapper(*v)
        return v

    def __iter__(self):
        'iterate over sequences in this alignment'
        for seqID in self.seqIDdict:
            yield self.nlmsa.seqDict[seqID]

    def getSeqID(self, seq):
        'return fully qualified sequence ID for this seq'
        return (~(self.nlmsa.seqDict))[seq]

    def __setitem__(self, k, ns):
        'save mapping of seq to the specified NLMSASequence'
        self.seqlist.append(ns)
        if isinstance(k, types.StringType):
            # Allow build with a string object.
            self._cache[k] = (ns.id, ns, 0)
        elif k is not None:
            self._cache[k.pathForward] = (ns.id, ns, 0)

    def __iadd__(self, ns):
        'add coord system ns to the alignment'
        self[None] = ns
        return self # iadd must return self!!!

    def close(self):
        'finalize and close shelve indexes'
        try:
            do_close = self.seqIDdict.close
        except AttributeError:
            return # our storage doesn't support close(), so nothing to do
        do_close() # close both shelve objects
        self.IDdict.close()

    def reopenReadOnly(self, mode='r'):
        'save existing data and reopen in read-only mode'
        self.close()
        self.seqIDdict = classutil.open_shelve(self.filename + '.seqIDdict',
                                               mode)
        self.IDdict = classutil.open_shelve(self.filename + '.idDict', mode)

    def getUnionSlice(self, seq):
        'get union coords for this seq interval, adding seq to index if needed'
        try:
            id, ns, offset = self[seq] # look up in index
        except KeyError:
            self.saveSeq(seq) # add this new sequence to our current union
            id, ns, offset = self[seq] # look up in index
        # Make sure to handle annotations right
        i, start, stop = self.getIDcoords(seq)
        if start < 0: # reverse orientation
            return ns, slice(start - offset, stop - offset) # use union coords
        else: # forward orientation
            return ns, slice(start + offset, stop + offset) # use union coords

    def clear_cache(self):
        'Clear the cache of saved sequences.'
        self._cache.clear()


def splitLPOintervals(lpoList, ival, targetIval=None):
    'return list of intervals split to different LPOs'
    if ival.start < 0: # reverse orientation: force into forward ori
        start= -(ival.stop)
        stop= -(ival.start)
    else: # forward orientation
        start=ival.start
        stop=ival.stop
    l = []
    i = len(lpoList) - 1
    while i >= 0:
        offset = lpoList[i].offset
        if offset < stop: # appears to be in this
            if offset <= start: # fits completely in this LPO
                if ival.start < 0: # reverse ori
                    myslice = slice(offset - stop, offset - start)
                else: # forward ori
                    myslice = slice(start - offset, stop - offset)
                if targetIval is not None:
                    l.append((lpoList[i], myslice, targetIval))
                else:
                    l.append((lpoList[i], myslice))
                return l # done
            else: # continues past start of this LPO
                if ival.start < 0: # reverse ori
                    myslice = slice(offset - stop, 0)
                else: # forward ori
                    myslice = slice(0, stop - offset)
                if targetIval is not None:
                    l.append((lpoList[i], myslice, targetIval[offset
                                                              - start:]))
                    # Remove the already-appended part
                    targetIval = targetIval[:offset - start]
                else:
                    l.append((lpoList[i], myslice))
                stop = offset
        i -= 1 # continue to previous LPO
    raise ValueError('empty lpoList or offset not starting at 0?  Debug!')


class BuildMSASlice(object):

    def __init__(self, ns, start, stop, id, offset, is_lpo=0, seq=None):
        self.ns = ns
        self.start = start
        self.stop = stop
        self.id = id
        self.offset = offset
        self.is_lpo = is_lpo
        self.seq = seq

    def offsetSlice(self, ival):
        if ival.orientation < 0:
            return slice(ival.start - self.offset, ival.stop - self.offset)
        else:
            return slice(ival.start + self.offset, ival.stop + self.offset)

    def __iadd__(self, targetIval):
        'save an alignment edge between self and targetIval'
        if self.is_lpo: # assign to correct LPO(s)
            if isinstance(targetIval, types.SliceType):
                raise ValueError('you attempted to map LPO --> LPO?!?')
            self.ns.nlmsaLetters.__iadd__(targetIval)
            splitList = splitLPOintervals(self.ns.nlmsaLetters.lpoList,
                                          slice(self.start, self.stop),
                                          targetIval)
            for ns, src, target in splitList:
                # Save intervals to respective LPOs; LPO --> target
                ns[src] = self.ns.nlmsaLetters.seqs.getIDcoords(target)
                if self.ns.nlmsaLetters.is_bidirectional:
                    nsu, myslice = self.ns.nlmsaLetters.seqs \
                            .getUnionSlice(target)
                    # Save target --> LPO
                    nsu[myslice] = (ns.id, src.start, src.stop)
        else:
            if isinstance(targetIval, types.SliceType): # target is LPO
                splitList = splitLPOintervals(self.ns.nlmsaLetters.lpoList,
                                            targetIval, self.seq)
                for ns, target, src in splitList:
                    self.ns[self.offsetSlice(src)] = (ns.id, target.start,
                                                      target.stop)
                    if self.ns.nlmsaLetters.is_bidirectional:
                        # Save LPO --> SRC
                        ns[target]=(self.id, src.start, src.stop)
            else: # both src and target are normal seqs.  use_virtual_lpo!!
                self.ns.nlmsaLetters.__iadd__(targetIval)
                self.ns.nlmsaLetters.init_pairwise_mode()
                # Our virtual LPO
                ns_lpo = self.ns.nlmsaLetters.seqlist[self.ns.id - 1]
                # Save src --> target
                ns_lpo[self.offsetSlice(self.seq)] = self.ns.nlmsaLetters \
                        .seqs.getIDcoords(targetIval)
                if self.ns.nlmsaLetters.is_bidirectional:
                    nsu, myslice = self.ns.nlmsaLetters.seqs \
                            .getUnionSlice(targetIval)
                    # Our virtual LPO
                    ns_lpo = self.ns.nlmsaLetters.seqlist[nsu.id - 1]
                    # Save target --> src
                    ns_lpo[myslice] = (self.id, self.start, self.stop)
        return self # iadd must always return self

    def __setitem__(self, k, v):
        if v is not None:
            raise ValueError('NLMSA cannot save edge-info. Only \
                             nlmsa[s1][s2]=None allowed')
        self += k


def read_seq_dict(pathstem, trypath=None):
    'read seqDict for NLMSA'
    if os.access(pathstem + '.seqDictP', os.R_OK):
        from pygr import worldbase
        ifile = file(pathstem+'.seqDictP', 'rb') # pickle is binary file!
        try: # load from worldbase-aware pickle file
            seqDict = worldbase._mdb.loads(ifile.read())
        finally:
            ifile.close()
    elif os.access(pathstem + '.seqDict', os.R_OK): # old-style union header
        import seqdb
        seqDict = seqdb.PrefixUnionDict(filename=pathstem+'.seqDict',
                                        trypath=trypath)
    else:
        raise ValueError('''Unable to find seqDict file
%s.seqDictP or %s.seqDict
and no seqDict provided as an argument''' % (pathstem, pathstem))
    return seqDict


def save_seq_dict(pathstem, seqDict):
    'save seqDict to a worldbase-aware pickle file'
    from metabase import dumps
    ofile = file(pathstem + '.seqDictP', 'wb') # pickle is binary file!
    try:
        ofile.write(dumps(seqDict))
    finally:
        ofile.close()


def prune_self_mappings(src_prefix, dest_prefix, is_bidirectional):
    '''return is_bidirectional flag according to whether source and
    target are the same genome.  This handles axtNet reading, in which
    mappings between genomes are given in only one direction, whereas
    mappings between the same genome are given in both directions.'''
    if src_prefix == dest_prefix:
        return 0
    else:
        return 1


def nlmsa_textdump_unpickler(filepath, kwargs):
    from cnestedlist import textfile_to_binaries, NLMSA
    logger.info('Saving NLMSA indexes from textdump: %s' % filepath)
    try:
        buildpath = os.environ['WORLDBASEBUILDDIR']
    except KeyError:
        buildpath = classutil.get_env_or_cwd('PYGRDATABUILDDIR')
    path = textfile_to_binaries(filepath, buildpath=buildpath, **kwargs)
    o = NLMSA(path) # now open in read mode from the saved index fileset
    o._saveLocalBuild = True # mark this for saving in local metabase
    return o


nlmsa_textdump_unpickler.__safe_for_unpickling__ = 1


class NLMSABuilder(object):
    'when unpickled triggers construction of NLMSA from textdump'
    _worldbase_no_cache = True # force worldbase to reload this fresh

    def __init__(self, filepath, **kwargs):
        self.filepath = filepath
        self.kwargs = kwargs

    def __reduce__(self):
        return (nlmsa_textdump_unpickler, (self.filepath, self.kwargs))


class SeqCacheOwner(object):
    'weak referenceable object: workaround for pyrex extension classes'

    def __init__(self):
        self.cachedSeqs = {}

    def cache_reference(self, seq):
        'keep a ref to seqs cached on our behalf'
        self.cachedSeqs[seq.id] = seq


def generate_nlmsa_edges(self, *args, **kwargs):
    """iterate over all edges for all sequences in the alignment.
    Very slow for a big alignment!"""
    for seq in self.seqs:
        myslice = self[seq]
        for results in myslice.edges(*args, **kwargs):
            yield results


def get_interval(seq, start, end, ori):
    "trivial function to get the interval seq[start:end] with requested ori"
    if ori < 0:
        return seq.absolute_slice(-end, -start)
    else:
        return seq.absolute_slice(start, end)


_default_ivals_attrs = dict(idDest='id', startDest='start',
                            stopDest='stop', oriDest='ori')


class CoordsToIntervals(object):
    '''Transforms coord objects to (ival1,ival2) aligned interval pairs.

    The intervals can come in in two forms:
    First, as a list, with [src, dest1, dest2, dest3] information;
    or second, as an object, with attributes specifying src/dest info.
    '''

    def __init__(self, srcDB, destDB=None,
                 alignedIvalsAttrs=_default_ivals_attrs):
        self.srcDB = srcDB
        if destDB:
            self.destDB = destDB
        else:
            self.destDB = srcDB
        self.getAttr = classutil.make_attribute_interface(alignedIvalsAttrs)

    def __call__(self, alignedCoords):
        '''Read interval info from alignedCoords and generate actual intervals.

        Information read is id, start, stop, and orientation (ori).
        '''
        for c in alignedCoords:
            if isinstance(c, (CoordsGroupStart, CoordsGroupEnd)):
                yield c # just pass grouping-info through
                continue

            try:
                srcData = c[0] # align everything to the first interval
                destSet = c[1:]
            except TypeError:
                srcData = c # extract both src and dest from ivals object
                destSet = [c]

            id = self.getAttr(srcData, 'id')
            start = self.getAttr(srcData, 'start')
            stop = self.getAttr(srcData, 'stop')
            ori = self.getAttr(srcData, 'ori', 1)    # default orientation: +

            srcIval = get_interval(self.srcDB[id], start, stop, ori)

            # get the dest interval(s) and yield w/src.
            for destData in destSet:
                idDest = self.getAttr(destData, 'idDest')
                startDest = self.getAttr(destData, 'startDest')
                stopDest = self.getAttr(destData, 'stopDest')
                oriDest = self.getAttr(destData, 'oriDest', 1) # default ori: +

                destIval = get_interval(self.destDB[idDest], startDest,
                                        stopDest, oriDest)

                yield srcIval, destIval # generate aligned intervals


def add_aligned_intervals(al, alignedIvals):
    '''Save a set of aligned intervals to alignment.
    '''
    # for each pair of aligned intervals, save them into the alignment.
    for t in alignedIvals:
        # is 't' a marker object for start or end of a group of coordinates?
        if isinstance(t, (CoordsGroupStart, CoordsGroupEnd)):
            continue # ignore grouping markers

        (src, dest) = t
        al += src
        al[src][dest] = None                # save their alignment


class CoordsGroupStart(object):
    '''Marker object indicating start of a coordinates group.

    See BlastHitParser for an example.'''
    pass


class CoordsGroupEnd(object):
    '''Marker object indicating end of a group of coordinates.

    See BlastHitParser for an example.'''
    pass
